import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import pyplot as plt
from dateutil.relativedelta import relativedelta

from pyproj import Proj, Transformer

import pylenm2
from pylenm2.data import filters, fetchers, transformation
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

timeseries_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def plot_all_time_series_simple(
        # self, 
        data_pylenm_dm,
        analyte_name=None, 
        start_date=None, 
        end_date=None, 
        title='Dataset: Time ranges', 
        x_label='Station', 
        y_label='Year',
        min_days=10, 
        x_min_lim=-5, 
        x_max_lim = 170, 
        y_min_date='1988-01-01', 
        y_max_date='2020-01-01', 
        return_data=False, 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Plots the start and end date of analyte readings for differnt locations/sensors/stations.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analyte_name (str, optional): analyte to examine. Defaults to None.
        start_date (str, optional): start date of horizontal time to show alignment. Defaults to None.
        end_date (str, optional): end date of horizontal time to show alignment.. Defaults to None.
        title (str, optional): plot title. Defaults to 'Dataset: Time ranges'.
        x_label (str, optional): x axis label. Defaults to 'Station'.
        y_label (str, optional): y axis label. Defaults to 'Year'.
        min_days (int, optional): minimum number of days required to plot the time series . Defaults to 10.
        x_min_lim (int, optional): x axis starting point. Defaults to -5.
        x_max_lim (int, optional): x axis ending point. Defaults to 170.
        y_min_date (str, optional): y axis starting date. Defaults to '1988-01-01'.
        y_max_date (str, optional): y axis ending date. Defaults to '2020-01-01'.
        return_data (bool, optional): flag to return data. Defaults to False.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
    """
    # data = self.simplify_data()
    data = filters.simplify_data(data=data_pylenm_dm)
    
    # Filtering
    # -----------------
    if(filter):
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.construction_data, 
            col=col, 
            equals=equals,
        )
        if('ERROR:' in str(filter_res)):
            timeseries_logger.error("Ran into ERROR when calling filter_by_column()!")
            return filter_res
        
        query_stations = list(data.STATION_ID.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))
        if(len(intersect_stations)<=0):
            timeseries_logger.error('ERROR: No results for this query with the specifed filter parameters.')
            return 'ERROR: No results for this query with the specifed filter parameters.'
        
        data = data[data['STATION_ID'].isin(intersect_stations)]

    # Preparing data
    # -----------------
    if(analyte_name!=None):
        data = data[data.ANALYTE_NAME == analyte_name]
    
    stations = data.STATION_ID.unique()
    stations_dateRange=pd.DataFrame(columns=['STATION_ID','START_DATE','END_DATE'])
    for i in range(len(stations)):
        stationName=stations[i]
        stationNamedData=data[data['STATION_ID']==stations[i]]
        minDate=min(stationNamedData['COLLECTION_DATE'])
        maxDate=max(stationNamedData['COLLECTION_DATE'])
        stations_dateRange.loc[stations_dateRange.shape[0]]=[stationName,minDate,maxDate]

    stations_dateRange["RANGE"] = stations_dateRange.END_DATE - stations_dateRange.START_DATE
    
    # stations_dateRange.RANGE = stations_dateRange.RANGE.astype('timedelta64[D]').astype('int')
    # stations_dateRange = stations_dateRange[stations_dateRange.RANGE>min_days]
    stations_dateRange = stations_dateRange[stations_dateRange.RANGE.dt.days>min_days]
    
    stations_dateRange.sort_values(by=["RANGE","END_DATE","START_DATE"], ascending = (False, False, True), inplace=True)
    stations_dateRange.reset_index(inplace=True)
    stations_dateRange.drop('index', axis=1, inplace=True)
    stations = np.array(stations_dateRange.STATION_ID)

    # Plotting
    # -----------------
    fig, ax = plt.subplots(1, 1, sharex=False,figsize=(20,6),dpi=300)

    ax.set_xticks(range(len(stations)))
    ax.set_xticklabels(stations, rotation='vertical', fontsize=6)

    ax.plot(stations_dateRange['START_DATE'], c='blue', marker='o',lw=0, label='Start date')
    ax.plot(stations_dateRange['END_DATE'], c='red', marker='o',lw=0, label='End date')

    ax.hlines([max(stations_dateRange['END_DATE'])], x_min_lim, x_max_lim, colors='purple', label='Selected end date')
    if(start_date==None):
        ax.hlines([min(stations_dateRange['START_DATE'])], x_min_lim, x_max_lim, colors='green', label='Selected start date')
    else:
        ax.hlines([pd.to_datetime(start_date)], x_min_lim, x_max_lim, colors='green', label='Selected start date')

    x_label = x_label + ' (count: ' + str(stations_dateRange.shape[0])+ ')'
    ax.set_xlabel(x_label, fontsize=20)
    ax.set_ylabel(y_label, fontsize=20)   
    ax.set_xlim([x_min_lim, x_max_lim])
    ax.set_ylim([pd.to_datetime(y_min_date), pd.to_datetime(y_max_date)]) 
    ax.plot([], [], ' ', label="Time series with at least {} days".format(min_days))
    ax.legend()
    
    if(analyte_name!=None):
        title = title + ' (' + analyte_name + ')'
    fig.suptitle(title, fontsize=20)
    for i in range(stations_dateRange.shape[0]):
        ax.vlines(i,stations_dateRange.loc[i,'START_DATE'],stations_dateRange.loc[i,'END_DATE'],colors='k')
    
    if(return_data):
        return stations_dateRange


def plot_all_time_series(
        # self, 
        data_pylenm_dm,
        analyte_name=None, 
        title='Dataset: Time ranges', 
        x_label='Station', 
        y_label='Year', 
        x_label_size=8, 
        marker_size=30,
        min_days=10, 
        x_min_lim=None, 
        x_max_lim=None, 
        y_min_date=None, 
        y_max_date=None, 
        sort_by_distance=True, 
        source_coordinate=[436642.70,3681927.09], 
        log_transform=False, 
        cmap=mpl.cm.rainbow, 
        drop_cols=[], 
        return_data=False, 
        filter=False, 
        col=None, 
        equals=[], 
        cbar_min=None, 
        cbar_max=None, 
        reverse_y_axis=False, 
        fontsize = 20, 
        figsize=(20,6), 
        dpi=300, 
        y_2nd_label=None,
    ):      # TODO: Getting error in the notebook!
    """Plots the start and end date of analyte readings for differnt locations/sensors/stations with colored concentration reading.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analyte_name (str, optional): analyte to examine. Defaults to None.
        title (str, optional): plot title. Defaults to 'Dataset: Time ranges'.
        x_label (str, optional): x axis label. Defaults to 'Station'.
        y_label (str, optional): y axis label. Defaults to 'Year'.
        x_label_size (int, optional): x axis label font size. Defaults to 8.
        marker_size (int, optional): point size for time series. Defaults to 30.
        min_days (int, optional): minimum number of days required to plot the time series . Defaults to 10.
        x_min_lim (int, optional): x axis starting point. Defaults to None.
        x_max_lim (int, optional): x axis ending point. Defaults to None.
        y_min_date (str, optional): y axis starting date. Defaults to None.
        y_max_date (str, optional): y axis ending date. Defaults to None.
        sort_by_distance (bool, optional): flag to sort by distance from source center. Defaults to True.
        source_coordinate (list, optional): Easting, Northing coordinate of source center. Defaults to [436642.70,3681927.09].
        log_transform (bool, optional): flag to toggle log base 10 transformation. Defaults to False.
        cmap (cmap, optional): color map for plotting. Defaults to mpl.cm.rainbow.
        drop_cols (list, optional): columns, usually stations, to exclude. Defaults to [].
        return_data (bool, optional): flag to return data. Defaults to False.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
        cbar_min (float, optional): color bar lower boundary. Defaults to None.
        cbar_max (float, optional): color bar upper boundary. Defaults to None.
        reverse_y_axis (bool, optional): flag that reverses y axis. Defaults to False.
        fontsize (int, optional): plot font size. Defaults to 20.
        figsize (tuple, optional): matplotlib style figure size. Defaults to (20,6).
        dpi (int, optional): DPI of figure. Defaults to 300.
        y_2nd_label (str, optional): color bar label manual override. Defaults to None.
    """
    
    # Get clean data
    # dt = self.getCleanData([analyte_name])
    dt = fetchers.getCleanData(
        data_pylenm_dm, 
        analytes=[] if analyte_name is None else [analyte_name],
    )
    dt = dt[analyte_name]
    
    # try:
    # Filter data (if necessary)
    if(filter):
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.get_construction_data(), 
            col=col, 
            equals=equals,
        )
        if('ERROR:' in str(filter_res)):
            timeseries_logger.error("Ran into ERROR when calling filter_by_column()!")
            return filter_res
        
        query_stations = list(dt.columns.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations) & set(dt.columns))
        if(len(intersect_stations)<=0):
            timeseries_logger.error('ERROR: No results for this query with the specifed filter parameters.')
            return 'ERROR: No results for this query with the specifed filter parameters.'
        dt = dt[intersect_stations]

    # Get station information
    station_info = data_pylenm_dm.get_construction_data()
    shared_stations = list(set(station_info.index) & set(dt.columns))
    dt = dt[shared_stations]
    station_info = station_info.T[shared_stations]
    dt = dt.reindex(sorted(dt.columns), axis=1)
    station_info = station_info.reindex(sorted(station_info.columns), axis=1)
    station_info = station_info.T
    
    # Add distance to source
    transformer = Transformer.from_crs("epsg:4326", "epsg:26917") # Latitude/Longitude to UTM
    UTM_x, UTM_y = transformer.transform(station_info.LATITUDE, station_info.LONGITUDE)
    X = np.vstack((UTM_x,UTM_y)).T
    station_info = pd.DataFrame(X, index=list(station_info.index),columns=['Easting', 'Northing'])
    station_info = transformation.add_dist_to_source(
        XX=station_info, 
        source_coordinate=source_coordinate,
    )

    # Sort by distance (if necessary)
    if(sort_by_distance):
        station_info.sort_values(
            by=['dist_to_source'], 
            ascending=True, 
            inplace=True,
        )
    
    dt = dt[station_info.index]

    # except Exception as e:
    #     timeseries_logger.error(f"Error occurred but ignoring the error - {e}")

    # Iterpolate/drop data
    dt = dt.interpolate()
    dt = dt.drop(drop_cols, axis=1) # DROP BAD ONES 
    
    # Take log transform of the data (if necessary)
    if(log_transform):
        dt[dt <= 0] = 0.00000001
        dt = np.log10(dt)
    
    # Plot data
    stations = dt.columns
    # if(cbar_min==None):
    #     cbar_min = dt.min().min()
    # if(cbar_max==None):
    #     cbar_max = dt.max().max()
    cbar_min = dt.min().min() if cbar_min is None else cbar_min 
    cbar_max = dt.max().max() if cbar_max is None else cbar_max
    norm = mpl.colors.Normalize(vmin=cbar_min, vmax=cbar_max)

    fig, ax = plt.subplots(
        1, 2, sharex=False, figsize=figsize, dpi=dpi, 
        gridspec_kw={'width_ratios': [40, 1]},
    )
    ax[0].set_xticks(range(len(stations)))
    ax[0].set_xticklabels(stations, rotation='vertical', fontsize=x_label_size)

    for col in stations:
        curr_start = dt[col].first_valid_index()
        curr_end =  dt[col].last_valid_index()
        length = len(list(dt[col].loc[curr_start:curr_end].index))
        color_vals = list(dt[col].loc[curr_start:curr_end])
        color_vals = [cmap(norm((x))) for x in color_vals]
        
        x = col
        ys = list(dt[col].loc[curr_start:curr_end].index)
        ax[0].scatter([str(x)]*length, ys, c=color_vals, marker='o',lw=0,s=marker_size, alpha=0.75)

    # Set plot labels
    x_label = x_label + ' (count: ' + str(dt.shape[1])+ ')'
    ax[0].set_xlabel(x_label, fontsize=fontsize)
    ax[0].set_ylabel(y_label, fontsize=fontsize)
    
    # Set plot limits
    # if(x_min_lim==None):
    #     x_min_lim = -5
    # if(x_max_lim==None):
    #     x_max_lim = len(stations)+5
    x_min_lim = -5 if x_min_lim is None else x_min_lim
    x_max_lim = len(stations)+5 if x_max_lim is None else x_max_lim
    ax[0].set_xlim([x_min_lim, x_max_lim])
    
    if (y_min_date==None):
        y_min_date = dt.index.min() + relativedelta(years=-1)
    if (y_max_date==None):
        y_max_date = dt.index.max() + relativedelta(years=1)
    ax[0].set_ylim([pd.to_datetime(y_min_date), pd.to_datetime(y_max_date)]) 
    ax[0].plot([], [], ' ', label="Time series with at least {} days".format(min_days))
    ax[0].set_facecolor((0, 0, 0, 0.1))
            
    # Add colorbar
    label_cb = "Concentration ({})".format(data_pylenm_dm.get_unit(analyte_name))
    if(log_transform):
        label_cb = "Log " + label_cb
    cbar = fig.colorbar(
        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
        cax=ax[1], 
        orientation='vertical',
    )
    if(y_2nd_label!=None):
        label_cb = y_2nd_label
    cbar.set_label(label=label_cb, size=fontsize)

    ax[0].tick_params(axis='y', labelsize=fontsize)
    cbar.ax.tick_params(labelsize=fontsize)
    plt.tight_layout()
    if(reverse_y_axis):
        ax[0].invert_yaxis()
    
    if(analyte_name!=None):
        title = title + ' (' + analyte_name + ')'
    fig.suptitle(title, fontsize=fontsize, y=1.05)
    
    # Return data if necessary
    if(return_data):
        return dt



from ipywidgets import HTML
from ipyleaflet import (Map, basemaps, Marker,FullScreenControl, 
                        Popup, AwesomeIcon) 


def plot_coordinates_to_map(
        # self, 
        gps_data, 
        center=[33.271459, -81.675873], 
        zoom=14,
    ) -> Map:
    """Plots the station locations on an interactive map given coordinates.

    Args:
        gps_data (pd.DataFrame): Data frame with the following column names: station_id, latitude, longitude, color. If the color column is not passed, the default color will be blue.
        center (list, optional): latitude and longitude coordinates to center the map view. Defaults to [33.271459, -81.675873].
        zoom (int, optional): value to determine the initial scale of the map. Defaults to 14.

    Returns:
        ipyleaflet.Map
    """

    # center = center
    # zoom = 14

    # Create the basemap
    m = Map(
        basemap=basemaps.Esri.WorldImagery, 
        center=center, 
        zoom=zoom,
    )

    m.add_control(FullScreenControl())
    for (index,row) in gps_data.iterrows():

        # Create icons for stations locations
        if('color' in gps_data.columns):
            icon = AwesomeIcon(
                name='tint',
                marker_color=row.loc['color'],
                icon_color='black',
                spin=False
            )
        else:
            icon = AwesomeIcon(
                name='tint',
                marker_color='blue',
                icon_color='black',
                spin=False
            )

        loc = [row.loc['LATITUDE'],row.loc['LONGITUDE']]
        station = HTML(value=row.loc['STATION_ID'])

        # Create the Popup
        popup = Popup(
            location=loc,
            child=station,
            close_button=True,
            auto_close=False,
            max_height=1,
            close_on_escape_key=False,
        )
        
        # Create the Marker and add the Popup to it
        marker = Marker(location=loc,
                        icon=icon,
                        draggable=False,
                        popup=popup,
                    )

        m.add_layer(marker)


        # marker.popup = popup

    return m




import os
import numpy as np
import pandas as pd
from tqdm import tqdm

from pylenm2.data import filters
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

fetchers_logger = logger_config.setup_logging(
    module_name=__name__,
    level=logging.INFO,
    # level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def _get_individual_analyte_df(data, dates, analyte):
    """<Function docstring> TODO: write function docstring

    Args:
        data (pd.DataaFrame): input dataframe.
        dates (pd.Series): input dates.
            TODO: Confirm the data type.
        analyte (str): name of the analyte.
    
    Returns:
        pd.DataFrame: Sample analytes based on the input arguments.
    """

    sample = data[analyte]
    sample_analyte = pd.DataFrame(sample, index=dates, columns=sample.index)
    
    for station in sample.index:
        sample_analyte[station] = sample[station]
    
    return sample_analyte


# Helper function to return start and end date for a date and a lag (+/- days)
# def __getLagDate(self, date, lagDays=7):
def _getLagDate(date, lagDays=7):
    date = pd.to_datetime(date)
    dateStart = date - pd.DateOffset(days=lagDays)
    dateEnd = date + pd.DateOffset(days=lagDays)
    return dateStart, dateEnd


def getCleanData(data_pylenm_dm, analytes):
    """Creates a table filling the data from the concentration dataset for a given analyte list where the columns are multi-indexed as follows [analytes, station names] and the index is all of the dates in the dataset. Many NaN should be expected.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analytes (list): list of analyte names to use

    Returns:
        pd.DataFrame
    """
    curr = data_pylenm_dm.data[['STATION_ID', 'COLLECTION_DATE', 'ANALYTE_NAME', 'RESULT']]
    
    # Get data from only the specified analytes (TODO: improve code)
    main = pd.DataFrame()
    for ana in analytes:
        main = pd.concat([main, curr[curr.ANALYTE_NAME==ana]])

    # Get the pivot data results
    piv = main.pivot_table(index=['COLLECTION_DATE'],columns=['ANALYTE_NAME', 'STATION_ID'], values='RESULT', aggfunc=np.mean)
    
    piv.index = pd.to_datetime(piv.index)
    piv.sort_index(inplace=True)
    
    return piv


def getCommonDates(
        # self, 
        data_pylenm_dm,
        analytes, 
        lag=[3,7,10],
    ) -> pd.DataFrame:
    """Creates a table which counts the number of stations within a range specified by a list of lag days.
    TODO: Very slow. Check if efficiency can be improved.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analytes (list): list of analyte names to use
        lag (list, optional): list of days to look ahead and behind the specified date (+/-). Defaults to [3,7,10].

    Returns:
        pd.DataFrame
    """
    # piv = self.getCleanData(analytes)
    piv = getCleanData(
        data_pylenm_dm=data_pylenm_dm,
        analytes=analytes,
    )
    dates = piv.index
    names=['Dates', 'Lag']
    tuples = [dates, lag]
    finalData = pd.DataFrame(index=pd.MultiIndex.from_product(tuples, names=names), columns=['Date Ranges', 'Number of stations'])
    
    for date in dates:
        for i in lag:
            # dateStart, dateEnd = self.__getLagDate(date, lagDays=i)
            dateStart, dateEnd = _getLagDate(date, lagDays=i)
            mask = (piv.index > dateStart) & (piv.index <= dateEnd)
            result = piv[mask].dropna(axis=1, how='all')
            numStations = len(list(result.columns.get_level_values(1).unique()))
            dateRange = str(dateStart.date()) + " - " + str(dateEnd.date())
            finalData.loc[date, i]['Date Ranges'] = dateRange
            finalData.loc[date, i]['Number of stations'] = numStations
    
    return finalData


def getJointData(
        data_pylenm_dm, 
        analytes, 
        lag=3,
    ) -> pd.DataFrame:
    """Creates a table filling the data from the concentration dataset for a given analyte list where the columns are multi-indexed as follows [analytes, station names] and the index is the date ranges secified by the lag.

    TODO: Replace current progress implementation with tqdm.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analytes (list): list of analyte names to use
        lag (int, optional): number of days to look ahead and behind the specified date (+/-). Defaults to 3.

    Returns:
        pd.DataFrame
    """
    # if(self.jointData_is_set(lag=lag)==True):
    if(data_pylenm_dm.is_set_jointData(lag=lag)):
        finalData = data_pylenm_dm.__jointData[0]
        return finalData
    
    piv = getCleanData(
        data_pylenm_dm=data_pylenm_dm,
        analytes=analytes,
    )
    
    dates = piv.index
    dateRanges = []
    for date in dates:
        dateStart, dateEnd = _getLagDate(date, lagDays=lag)
        dateRange = str(dateStart.date()) + " - " + str(dateEnd.date())
        dateRanges.append(dateRange)
    
    finalData = pd.DataFrame(columns=piv.columns, index=dateRanges)
    # numLoops = len(dates)
    # everySomePercent = []
    print("Generating data with a lag of {} days.".format(lag).upper())
    # print("Progress:")
    # for x in list(np.arange(1, 100, 1)):
    #     everySomePercent.append(round((x/100)*numLoops))
    
    # for date, iteration in zip(dates, range(numLoops)):
    for date in tqdm(dates, desc="Progress"):
        # if(iteration in everySomePercent):
        #     print(str(round(iteration/numLoops*100)) + "%", end=', ')
        
        dateStart, dateEnd = _getLagDate(date, lagDays=lag)
        dateRange = str(dateStart.date()) + " - " + str(dateEnd.date())
        mask = (piv.index > dateStart) & (piv.index <= dateEnd)
        result = piv[mask].dropna(axis=1, how='all')
        resultCollapse = pd.concat([result[col].dropna().reset_index(drop=True) for col in result], axis=1)
        
        # HANDLE MULTIPLE VALUES
        if(resultCollapse.shape[0]>1):
            resultCollapse = pd.DataFrame(resultCollapse.mean()).T
        resultCollapse = resultCollapse.rename(index={0: dateRange})
        for ana_station in resultCollapse.columns:
            finalData.loc[dateRange, ana_station] =  resultCollapse.loc[dateRange, ana_station]
    
        # Save data to the pylenm global variable
        # data_pylenm_dm.__set_jointData(data=finalData, lag=lag)
        data_pylenm_dm.set_jointData(data=finalData, lag=lag)
    
    for col in finalData.columns:
        finalData[col] = finalData[col].astype('float64')
    
    print("Completed")
    return finalData


def get_analyte_details(
        data_pylenm_dm, 
        analyte_name, 
        filter=False, 
        col=None, 
        equals=[], 
        save_to_file = False, 
        save_dir='analyte_details',
    ):
    """Returns a csv file saved to save_dir with details pertaining to the specified analyte. Details include the station names, the date ranges and the number of unique samples.

    TODO: Handle error returns better!

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analyte_name (str): name of the analyte to be processed
        filter (bool, optional): whether to filter the data. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
        save_to_file (bool, optional): whether to save data to file. Defaults to False.
        save_dir (str, optional): name of the directory you want to save the csv file to. Defaults to 'analyte_details'.

    Returns:
        pd.DataFrame: Table with station information
    """
    data = data_pylenm_dm.data
    data = data[data.ANALYTE_NAME == analyte_name].reset_index().drop('index', axis=1)
    data = data[~data.RESULT.isna()]
    data = data.drop(['ANALYTE_NAME', 'RESULT', 'RESULT_UNITS'], axis=1)
    data.COLLECTION_DATE = pd.to_datetime(data.COLLECTION_DATE)
    if(filter):
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.construction_data, 
            col=col, 
            equals=equals,
        )
        if('ERROR:' in str(filter_res)):    # TODO: Handle Error returns better
            fetchers_logger.error("Ran into ERROR when calling filter_by_column()!")
            return filter_res
        
        query_stations = list(data.STATION_ID.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))
        if(len(intersect_stations)<=0):
            fetchers_logger.error('ERROR: No results for this query with the specifed filter parameters.')
            return 'ERROR: No results for this query with the specifed filter parameters.'
        data = data[data['STATION_ID'].isin(intersect_stations)]        

    info = []
    stations = np.unique(data.STATION_ID.values)
    for station in stations:
        current = data[data.STATION_ID == station]
        startDate = current.COLLECTION_DATE.min().date()
        endDate = current.COLLECTION_DATE.max().date()
        numSamples = current.duplicated().value_counts()[0]
        info.append({
            'Station Name': station, 
            'Start Date': startDate, 
            'End Date': endDate,
            'Date Range (days)': endDate-startDate ,
            'Unique samples': numSamples,
        })
    details = pd.DataFrame(info)
    details.index = details['Station Name']
    details = details.drop('Station Name', axis=1)
    details = details.sort_values(by=['Start Date', 'End Date'])
    details['Date Range (days)'] = (details['Date Range (days)']/ np.timedelta64(1, 'D')).astype(int)
    if(save_to_file):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        details.to_csv(save_dir + '/' + analyte_name + '_details.csv')
    return details


def get_data_summary(
        data_pylenm_dm, 
        analytes=None, 
        sort_by='date', 
        ascending=False, 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Returns a dataframe with a summary of the data for certain analytes. Summary includes the date ranges and the number of unique samples and other statistics for the analyte results.

    TODO: Handle error returns better!

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analytes (list, optional): list of analyte names to be processed. If left empty, a list of all the analytes in the data will be used. Defaults to None.
        sort_by (str, optional): {‘date’, ‘samples’, ‘stations’} sorts the data by either the dates by entering: ‘date’, the samples by entering: ‘samples’, or by unique station locations by entering ‘stations’. Defaults to 'date'.
        ascending (bool, optional): flag to sort in ascending order.. Defaults to False.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].

    Returns:
        pd.DataFrame: Table with station information
    """
    data = data_pylenm_dm.data
    if(analytes == None):
        analytes = data.ANALYTE_NAME.unique()
    data = data.loc[data.ANALYTE_NAME.isin(analytes)].drop(['RESULT_UNITS'], axis=1)
    data = data[~data.duplicated()] # remove duplicates
    data.COLLECTION_DATE = pd.to_datetime(data.COLLECTION_DATE)
    data = data[~data.RESULT.isna()]
    if(filter):
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.construction_data, 
            col=col, 
            equals=equals,
        )
        if('ERROR:' in str(filter_res)):    # TODO: Handle Error returns better
            fetchers_logger.error("Ran into ERROR when calling filter_by_column()!")
            return filter_res
        
        query_stations = list(data.STATION_ID.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))
        if(len(intersect_stations)<=0):
            fetchers_logger.error('ERROR: No results for this query with the specifed filter parameters.')
            return 'ERROR: No results for this query with the specifed filter parameters.'
        data = data[data['STATION_ID'].isin(intersect_stations)]

    info = []
    for analyte_name in analytes:
        query = data[data.ANALYTE_NAME == analyte_name]
        startDate = min(query.COLLECTION_DATE)
        endDate = max(query.COLLECTION_DATE)
        numSamples = query.shape[0]
        stationCount = len(query.STATION_ID.unique())
        stats = query.RESULT.describe().drop('count', axis=0)
        stats = pd.DataFrame(stats).T
        stats_col = [x for x in stats.columns]

        result = {
            'Analyte Name': analyte_name, 
            'Start Date': startDate, 
            'End Date': endDate, 
            'Date Range (days)':endDate-startDate, 
            '# unique stations': stationCount, 
            '# samples': numSamples, 
            'Unit': data_pylenm_dm.get_unit(analyte_name), 
        }
        for num in range(len(stats_col)):
            result[stats_col[num]] = stats.iloc[0][num] 

        info.append(result)

    details = pd.DataFrame(info)
    details.index = details['Analyte Name']
    details = details.drop('Analyte Name', axis=1)
    if(sort_by.lower() == 'date'):
        details = details.sort_values(by=['Start Date', 'End Date', 'Date Range (days)'], ascending=ascending)
    elif(sort_by.lower() == 'samples'):
        details = details.sort_values(by=['# samples'], ascending=ascending)
    elif(sort_by.lower() == 'stations'):
        details = details.sort_values(by=['# unique stations'], ascending=ascending)

    return details


def get_station_analytes(
        data_pylenm_dm, 
        station_names=None, 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Displays the analyte names available at given station locations.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_names (str, List[str], optional): names of the station. If left empty, all stations are returned. The input can either be a str, List[str], or None. Defaults to None.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].

    Returns:
        None
    """
    data = data_pylenm_dm.data
    bb = "\033[1m"
    be = "\033[0m"
    
    if(filter):
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.construction_data, 
            col=col, 
            equals=equals,
        )
        
        if('ERROR:' in str(filter_res)):
            fetchers_logger.error("Ran into ERROR when calling filter_by_column()!")
            return filter_res
        
        query_stations = list(data.STATION_ID.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))
        if(len(intersect_stations)<=0):
            fetchers_logger.error('ERROR: No results for this query with the specifed filter parameters.')
            return 'ERROR: No results for this query with the specifed filter parameters.'
        data = data[data['STATION_ID'].isin(intersect_stations)]
    
    if(station_names==None):
        stations = list(data.STATION_ID.unique())
    elif isinstance(station_names, str):
        stations = [station_names]
    elif isinstance(station_names, list):
        if len(station_names) > 0:
            stations = station_names
        else:
            stations = list(data.STATION_ID.unique())
    else:
        fetchers_logger.error("Invalid input for station_names! The input should either be a str, List[str], or None.")
        raise ValueError("Invalid input for station_names! The input should either be a str, List[str], or None.")
    
    station_analytes = dict()
    for station in stations:
        analytes = sorted(list(data[data.STATION_ID==station].ANALYTE_NAME.unique()))
        
        fetchers_logger.debug(f"{bb}{str(station)}{be}: {str(analytes)}")

        station_analytes[station] = analytes

    return station_analytes


import os
import numpy as np
import pandas as pd
import seaborn as sns
from tqdm import tqdm
from matplotlib import pyplot as plt
from sklearn.preprocessing import StandardScaler

from pylenm2.stats import preprocess
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

heatmap_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def plot_correlation_heatmap(
        # self, 
        data_pylenm_dm, 
        station_name, 
        show_symmetry=True, 
        color=True, 
        save_dir='plot_correlation_heatmap',
    ):
    """ Plots a heatmap of the correlations of the important analytes over time for a specified station.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        show_symmetry (bool, optional): choose whether or not the heatmap should show the same information twice over the diagonal. Defaults to True.
        color (bool, optional): choose whether or not the plot should be in color or in greyscale. Defaults to True.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_correlation_heatmap'.

    Returns:
        None
    """
    
    data = data_pylenm_dm.data
    query = data[data.STATION_ID == station_name]
    
    a = list(np.unique(query.ANALYTE_NAME.values))
    b = ['TRITIUM','IODINE-129','SPECIFIC CONDUCTANCE', 'PH','URANIUM-238', 'DEPTH_TO_WATER']
    
    analytes = preprocess._custom_analyte_sort(list(set(a) and set(b)))
    query = query.loc[query.ANALYTE_NAME.isin(analytes)]
    analytes = preprocess._custom_analyte_sort(np.unique(query.ANALYTE_NAME.values))
    
    x = query[['COLLECTION_DATE', 'ANALYTE_NAME']]
    unique = ~x.duplicated()
    query = query[unique]
    
    piv = query.reset_index().pivot(
        index='COLLECTION_DATE', 
        columns='ANALYTE_NAME', 
        values='RESULT',
    )
    piv = piv[analytes]
    totalSamples = piv.shape[0]
    piv = piv.dropna()
    samples = piv.shape[0]
    
    if(samples < 5):
        heatmap_logger.error(f"ERROR: {station_name} does not have enough samples to plot.")
        return None
    
    else:
        scaler = StandardScaler()
        pivScaled = scaler.fit_transform(piv)
        pivScaled = pd.DataFrame(pivScaled, columns=piv.columns)
        pivScaled.index = piv.index
        
        corr_matrix = pivScaled.corr()
        
        if(show_symmetry):
            mask = None
        else:
            mask = np.triu(corr_matrix)
        if(color):
            cmap = 'RdBu'
        else:
            cmap = 'binary'
        
        fig, ax = plt.subplots(figsize=(8,6))
        ax.set_title(station_name + '_correlation', fontweight='bold')
        ttl = ax.title
        ttl.set_position([.5, 1.05])
        
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
        ax.text(
            1.3, 
            1.05, 
            # 'Start date: {}\nEnd date: {}\n\nSamples: {} of {}'.format(piv.index[0], piv.index[-1], samples, totalSamples), 
            f"Start date: {piv.index[0]}\nEnd date: {piv.index[-1]}\n\nSamples: {samples} of {totalSamples}", 
            transform=ax.transAxes, 
            fontsize=15, 
            fontweight='bold', 
            verticalalignment='bottom', 
            bbox=props,
        )
        
        ax = sns.heatmap(
            corr_matrix,
            ax=ax,
            mask=mask,
            vmin=-1, vmax=1,
            xticklabels=corr_matrix.columns,
            yticklabels=corr_matrix.columns,
            cmap=cmap,
            annot=True,
            linewidths=1,
            cbar_kws={'orientation': 'vertical'},
        )

        # Save plot
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        
        fig.savefig(
            # save_dir + '/' + station_name + '_correlation.png', 
            f"{save_dir}/{station_name}_correlation.png", 
            bbox_inches="tight",
        )

        # Returning a random non-None value to mark successful execution
        return 200


def plot_all_correlation_heatmap(
        # self, 
        data_pylenm_dm, 
        show_symmetry=True, 
        color=True, 
        save_dir='plot_correlation_heatmap',
    ):
    """Plots a heatmap of the correlations of the important analytes over time for each station in the dataset.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        show_symmetry (bool, optional): choose whether or not the heatmap should show the same information twice over the diagonal. Defaults to True.
        color (bool, optional): choose whether or not the plot should be in color or in greyscale. Defaults to True.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_correlation_heatmap'.
    """
    
    data = data_pylenm_dm.data
    
    stations = np.array(data.STATION_ID.values)
    stations = np.unique(stations)

    successful_plots = 0
    failed_plots = 0
    
    for station in tqdm(stations, desc="Stations"):
        plot_status = plot_correlation_heatmap(
            data_pylenm_dm=data_pylenm_dm,
            station_name=station,
            show_symmetry=show_symmetry,
            color=color,
            save_dir=save_dir,
        )

        if plot_status is None:
            failed_plots += 1
        else:
            successful_plots += 1
        
    heatmap_logger.info(f"Successful plots: {successful_plots}")
    heatmap_logger.info(f"Failed plots: {failed_plots}")


import os
import re
import numpy as np
import pandas as pd

import pylenm2
import pylenm2.data
import pylenm2.data.data_module
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

filters_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def simplify_data(
        data, 
        inplace=False, 
        columns=None, 
        save_csv=False, 
        file_name= 'data_simplified', 
        save_dir='data/',
    ):
    """Removes all columns except 'COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME', 'RESULT', and 'RESULT_UNITS', i.e. the `REQUIRED_DATA_COLUMNS`!
        
        If the user specifies additional columns in addition to the ones listed above, those columns will be kept.
        The function returns a dataframe and has an optional parameter to be able to save the dataframe to a csv file.

    Args:
        data (pd.DataFrame, pylenm2.PylenmDataModule, optional): data to simplify.
        inplace (bool, optional): save data to current working dataset. Defaults to False.
        columns (list, optional): list of any additional columns on top of  ['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME', 'RESULT', and 'RESULT_UNITS'] to be kept in the dataframe. Defaults to None.
        save_csv (bool, optional): flag to determine whether or not to save the dataframe to a csv file. Defaults to False.
        file_name (str, optional): name of the csv file you want to save. Defaults to 'data_simplified'.
        save_dir (str, optional): name of the directory you want to save the csv file to. Defaults to 'data/'.

    Returns:
        pd.DataFrame
    """
    # if(str(type(data)).lower().find('dataframe') == -1):
    #     data = self.data
    # else:
    #     data = data
    if isinstance(data, pd.DataFrame):
        data_df = data.copy(deep=True)
    elif isinstance(data, (pylenm2.PylenmDataModule, pylenm2.data.data_module.PylenmDataModule)):
        data_df = data.data.copy(deep=True)
    else:
        filters_logger.error("`data` must be either a pandas DataFrame or PylenmDataModule!")
        # raise ValueError("`data` must be either a pandas DataFrame or PylenmDataModule!")
        return None
        
    if columns==None:
        # sel_cols = ['COLLECTION_DATE','STATION_ID','ANALYTE_NAME','RESULT','RESULT_UNITS']
        sel_cols = c.REQUIRED_DATA_COLUMNS
    else:
        # hasColumns = all(item in list(data.columns) for item in columns)
        # if(hasColumns):

        if set(columns).issubset(data_df.columns):
            # sel_cols = ['COLLECTION_DATE','STATION_ID','ANALYTE_NAME','RESULT','RESULT_UNITS'] + columns
            sel_cols = c.REQUIRED_DATA_COLUMNS + columns
        else:
            extra_columns = set(columns).difference(data_df.columns)
            filters_logger.error(f'Following specified column(s) do not exist in the data: {extra_columns}')
            raise ValueError("Specified column(s) do not exist in the data!")

    data_df = data_df[sel_cols]
    data_df.COLLECTION_DATE = pd.to_datetime(data_df.COLLECTION_DATE, format=c.COLLECTION_DATE_FORMAT)
    data_df = data_df.sort_values(by="COLLECTION_DATE")
    dup = data_df[data_df.duplicated(['COLLECTION_DATE', 'STATION_ID','ANALYTE_NAME', 'RESULT'])]
    data_df = data_df.drop(dup.index)
    data_df = data_df.reset_index().drop('index', axis=1)
    
    if(save_csv):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        data_df.to_csv(save_dir + file_name + '.csv')
        print('Successfully saved "' + file_name +'.csv" in ' + save_dir)
    
    if(inplace):
        if isinstance(data, pd.DataFrame):
            data.drop(data.index, inplace=True)
            data.drop(columns=list(set(data.columns).difference(sel_cols)), inplace=True)
            data[sel_cols] = data_df[sel_cols]
    
        elif isinstance(data, (pylenm2.PylenmDataModule, pylenm2.data.data_module.PylenmDataModule)):
            data.set_data(data_df, verbose=False)
        
        else:
            raise pylenm2.UnreachableCodeError("Code execution should never reach here!")
    
    return data_df


def filter_by_column(data, col, equals=[]):
    """Filters construction data based on one column. You only specify ONE column to filter by, but can selected MANY values for the entry.

    TODO: Handle Error returns better!

    Args:
        data (pd.DataFrame): dataframe to filter.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].

    Returns:
        pd.DataFrame: returns filtered dataframe
    """
    if (data is None):
        filters_logger.error('ERROR: DataFrame was not provided to this function.')
        # return 'ERROR: DataFrame was not provided to this function.'
        return None
    else:
        # if(str(type(data)).lower().find('dataframe') == -1):
        if not isinstance(data, pd.DataFrame):
            filters_logger.error('ERROR: Data provided is not a pandas DataFrame.')
            # return 'ERROR: Data provided is not a pandas DataFrame.'
            return None
        else:
            data = data
    
    # DATA VALIDATION
    if (col==None):
        filters_logger.error('ERROR: Specify a column name to filter by.')
        # return 'ERROR: Specify a column name to filter by.'
        return None

    # data_cols = list(data.columns)
    # if((col in data_cols)==False): # Make sure column name exists 
    if col not in data.columns: # Make sure column name exists 
        filters_logger.error(f'Error: Column name {col} does not exist')
        # return 'Error: Column name "{}" does not exist'.format(col)
        return None
    
    if (equals==[]):
        filters_logger.error(f"ERROR: Specify a value that {col} should equal to.")
        # return 'ERROR: Specify a value that "{}" should equal to'.format(col)
        return None
    
    # data_val = list(data[col])
    # for value in equals:
    #     if((value in data_val)==False):
    #         return 'ERROR: No value equal to "{}" in "{}".'.format(value, col)
    values_not_in_data = set(equals).difference(set(data[col]))
    if len(values_not_in_data) > 0:
        filters_logger.error(f'ERROR: {values_not_in_data} do not exist in {col}.')
        # return f'ERROR: {values_not_in_data} do not exist in {col}.'
        return None

    # QUERY
    # final_data = pd.DataFrame()
    # for value in equals:
    #     current_data = data[data[col]==value]
    #     final_data = pd.concat([final_data, current_data])
    final_data = data[[item in equals for item in data[col]]]
    return final_data


def filter_stations(data_pylenm_dm, units):
    """Returns a list of the station names filtered by the unit(s) specified.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        units (list): Letter of the station to be filtered (e.g. [‘A’] or [‘A’, ‘D’])

    Returns:
        list: station names filtered by the unit(s) specified
    """
    
    data = data_pylenm_dm.data
    
    if units==None:
        units= ['A', 'B', 'C', 'D']
    elif not isinstance(units, (list, tuple)):
        units = [units]
    
    def getUnits():
        stations = list(np.unique(data.STATION_ID))
        stations = pd.DataFrame(stations, columns=['STATION_ID'])
        
        for index, row in stations.iterrows():
            mo = re.match('.+([0-9])[^0-9]*$', row.STATION_ID)
            last_index = mo.start(1)
            stations.at[index, 'unit'] = row.STATION_ID[last_index+1:]
            u = stations.unit.iloc[index]
            
            if(len(u)==0): # if has no letter, use D
                stations.at[index, 'unit'] = 'D'
            if(len(u)>1): # if has more than 1 letter, remove the extra letter
                if(u.find('R')>0):
                    stations.at[index, 'unit'] = u[:-1]
                else:
                    stations.at[index, 'unit'] = u[1:]
            
            u = stations.unit.iloc[index]
            
            if(u=='A' or u=='B' or u=='C' or u=='D'):
                pass
            else:
                stations.at[index, 'unit'] = 'D'
        return stations
    
    df = getUnits()
    
    res = df.loc[df.unit.isin(units)]
    
    return list(res.STATION_ID)


def query_data(data_pylenm_dm, station_name, analyte_name):
    """Filters data by passing the data and specifying the station_name and analyte_name

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analyte_name (str): name of the analyte to be processed

    Returns:
        pd.DataFrame: filtered data based on query conditons
    """
    data = data_pylenm_dm.data
    
    query = data[data.STATION_ID == station_name]
    query = query[query.ANALYTE_NAME == analyte_name]
    
    if query.shape[0]==0:
        return None
        # return 0        # TODO: Handle this better!
        # return pd.DataFrame(columns=data.columns)   # TODO: Use this once you make sure that the above return value is not being used anywhere else.
    else:
        return query



# Required imports
import os
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl
import pylab
import scipy
import random
import datetime
import re
import time
from math import sqrt
import matplotlib.dates as mdates
from matplotlib.dates import date2num, num2date
from sklearn import preprocessing
pd.set_option('display.max_columns', None) # to view all columns
from scipy.optimize import curve_fit
from supersmoother import SuperSmoother
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.pipeline import make_pipeline
from sklearn.gaussian_process.kernels import Matern, WhiteKernel
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression, Ridge, Lasso, RidgeCV, LassoCV
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import mean_squared_error
import scipy.stats as stats
import warnings
warnings.filterwarnings("ignore")
from pyproj import Proj, Transformer
from ipyleaflet import (Map, basemaps, WidgetControl, GeoJSON, 
                        LayersControl, Icon, Marker,FullScreenControl,
                        CircleMarker, Popup, AwesomeIcon) 
from ipywidgets import HTML
plt.rcParams["font.family"] = "Times New Roman"


import os
import sys
import time
import logging

from pylenm2.utils import constants as c


# def log_separator(logger, msg=f"\n\n\n{'-'*50}\n\n\n", *args, **kwargs):
#         # Get filehandler
#         original_fh_formatters = [(i, handler.formatter) for i, handler in enumerate(logger.handlers) if isinstance(handler, logging.FileHandler)]
             
#         # Log separators without any extra formatting other than the message
#         for i, orig_formatter in original_fh_formatters:
#              logger.handlers[i].setFormatter(logging.Formatter('%(message)s'))
#              logger.info(msg, *args, **kwargs)
#              logger.handlers[i].setFormatter(orig_formatter)



def setup_logging(
        module_name,
        level=None,
        logfile_dir=c.LOGFILE_DIR,
        get_stream_logs=True,
        main_module=False,
    ):

    # Make log directory if it doesn't exist
    os.makedirs(logfile_dir, exist_ok=True)

    # define default logging levels
    if level is None:
        default_level = logging.INFO
        filehandler_level = logging.DEBUG
        streamhandler_level = logging.ERROR if get_stream_logs else None
    else:
        default_level = level
        filehandler_level = level
        streamhandler_level = level if get_stream_logs else None

    # Create a logger object
    logger = logging.getLogger(module_name)
    logger.setLevel(default_level)  # logging.INFO by default

    if not logger.handlers:  # Avoid adding multiple handlers if already exist
        # Create file handler which logs messages in a file. One log file is used for each day.
        current_date = time.strftime("%Y-%m-%d", time.localtime())
        fh = logging.FileHandler(f"{logfile_dir}/{current_date}.log")
        fh.setLevel(filehandler_level)    # logging.DEBUG by default

        # Create console handler with a higher log level
        if get_stream_logs:
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(streamhandler_level)    # logging.ERROR by default

        # Create formatter and add it to the handlers
        formatter = logging.Formatter(c.LOG_FORMAT)
        fh.setFormatter(formatter)
        if get_stream_logs:
            ch.setFormatter(formatter)

        # Add the handlers to the logger
        logger.addHandler(fh)
        if get_stream_logs:
            logger.addHandler(ch)

    if main_module:
        logger.info(c.LOG_SEPARATOR)

    return logger


class PylenmDataFactory(object):
    def __init__(self, data: pd.DataFrame):
        ...
    def __isValid_Data(self, data):
        ...
    def __isValid_Construction_Data(self, data):
        ...
    def __hasColumns_Data(self, data):
        ...
    def __hasColumns_Construction_Data(self, data):
        ...
    def setData(self, data: pd.DataFrame, verbose: bool = True) -> None:
    def setConstructionData(self, construction_data: pd.DataFrame, verbose=True):
        ...
    def jointData_is_set(self, lag):
        ...
    def __set_jointData(self, data, lag):
        ...
    def getData(self):
        ...
    def get_Construction_Data(self):
        ...
    def __REQUIREMENTS_DATA(self):
        ...
    def __REQUIREMENTS_CONSTRUCTION_DATA(self):
        ...
    def __custom_analyte_sort(self, analytes):
        ...
    def __plotUpperHalf(self, *args, **kwargs):
        ...
    def simplify_data(self, data=None, inplace=False, columns=None, save_csv=False, file_name= 'data_simplified', save_dir='data/'):
        ...
    def get_MCL(self, analyte_name):
        ...
    def __set_units(self):
        ...
    def get_unit(self, analyte_name):
        ...
    def filter_by_column(self, data=None, col=None, equals=[]):
        ...
    def filter_wells(self, units):
        ...
    def remove_outliers(self, data, z_threshold=4):
        ...
    def get_analyte_details(self, analyte_name, filter=False, col=None, equals=[], save_to_file = False, save_dir='analyte_details'):
        ...
    def get_data_summary(self, analytes=None, sort_by='date', ascending=False, filter=False, col=None, equals=[]):
        ...
    def get_well_analytes(self, well_name=None, filter=False, col=None, equals=[]):
        ...
    def query_data(self, well_name, analyte_name):
        ...
    def plot_data(self, well_name, analyte_name, log_transform=True, alpha=0,
                plot_inline=True, year_interval=2, x_label='Years', y_label='', save_dir='plot_data', filter=False, col=None, equals=[]):
        ...

    def plot_all_data(self, log_transform=True, alpha=0, year_interval=2, plot_inline=True, save_dir='plot_data'):
        ...
    def plot_correlation_heatmap(self, well_name, show_symmetry=True, color=True, save_dir='plot_correlation_heatmap'):
        ...
    def plot_all_correlation_heatmap(self, show_symmetry=True, color=True, save_dir='plot_correlation_heatmap'):
        ...
    def interpolate_well_data(self, well_name, analytes, frequency='2W'):
        ...
    def plot_corr_by_well(self, well_name, analytes, remove_outliers=True, z_threshold=4, interpolate=False, frequency='2W', save_dir='plot_correlation', log_transform=False, fontsize=20, returnData=False, remove=[], no_log=None):
        ...
    def plot_all_corr_by_well(self, analytes, remove_outliers=True, z_threshold=4, interpolate=False, frequency='2W', save_dir='plot_correlation', log_transform=False, fontsize=20):
        ...
    def plot_corr_by_date_range(self, date, analytes, lag=0, min_samples=10, save_dir='plot_corr_by_date', log_transform=False, fontsize=20, returnData=False, no_log=None):
        ...
    def plot_corr_by_year(self, year, analytes, remove_outliers=True, z_threshold=4, min_samples=10, save_dir='plot_corr_by_year', log_transform=False, fontsize=20, returnData=False, no_log=None):
        ...
    def plot_MCL(self, well_name, analyte_name, year_interval=5, save_dir='plot_MCL'):
        ...
    def plot_PCA_by_date(self, date, analytes, lag=0, n_clusters=4, return_clusters=False, min_samples=3, show_labels=True, save_dir='plot_PCA_by_date', filter=False, col=None, equals=[]):
        ...
    def plot_PCA_by_year(self, year, analytes, n_clusters=4, return_clusters=False, min_samples=10, show_labels=True, save_dir='plot_PCA_by_year', filter=False, col=None, equals=[]):
        ...
    def plot_PCA_by_well(self, well_name, analytes, interpolate=False, frequency='2W', min_samples=10, show_labels=True, save_dir='plot_PCA_by_well'):
        ...
    def plot_coordinates_to_map(self, gps_data, center=[33.271459, -81.675873], zoom=14) -> Map:
    def interpolate_wells_by_analyte(self, analyte, frequency='2W', rm_outliers=True, z_threshold=3):
        ...
    def __transform_time_series(self, analytes=[], resample='2W', rm_outliers=False, z_threshold=4):
        ...
    def __get_individual_analyte_df(self, data, dates, analyte):
        ...
    def __cluster_data_OLD(self, data, n_clusters=4, log_transform=False, filter=False, filter_well_by=['D'], return_clusters=False):
        ...
    def cluster_data(self, data, analyte_name=["ANALYTE_NAME"], n_clusters=4, filter=False, col=None, equals=[], year_interval=5, y_label = 'Concentration', return_clusters=False ):
        ...
    def plot_all_time_series_simple(self, analyte_name=None, start_date=None, end_date=None, title='Dataset: Time ranges', x_label='Well', y_label='Year', min_days=10, x_min_lim=-5, x_max_lim = 170, y_min_date='1988-01-01', y_max_date='2020-01-01', return_data=False, filter=False, col=None, equals=[]):
        ...
    def plot_all_time_series(self, analyte_name=None, title='Dataset: Time ranges', x_label='Well', y_label='Year', x_label_size=8, marker_size=30, min_days=10, x_min_lim=None, x_max_lim=None, y_min_date=None, y_max_date=None, sort_by_distance=True, source_coordinate=[436642.70,3681927.09], log_transform=False, cmap=mpl.cm.rainbow, drop_cols=[], return_data=False, filter=False, col=None, equals=[], cbar_min=None, cbar_max=None, reverse_y_axis=False, fontsize = 20, figsize=(20,6), dpi=300, y_2nd_label=None):
        ...
    def __getLagDate(self, date, lagDays=7):
        ...
    def getCleanData(self, analytes):
        ...
    def getCommonDates(self, analytes, lag=[3,7,10]):
        ...
    def getJointData(self, analytes, lag=3):
        ...
    def mse(self, y_true, y_pred):
        ...
    def get_Best_GP(self, X, y, smooth=True, seed = 42):
        ...
    def fit_gp(self, X, y, xx, model=None, smooth=True):
        ...
    def interpolate_topo(self, X, y, xx, ft=['Elevation'], model=None, smooth=True, regression='linear', seed = 42):
        ...
    def __get_Best_Well(self, X, y, xx, ref, selected, leftover, ft=['Elevation'], regression='linear', verbose=True, smooth=True, model=None):
        ...
    def get_Best_Wells(self, X, y, xx, ref, initial, max_wells, ft=['Elevation'], regression='linear', verbose=True, smooth=True, model=None):
        ...
    def dist(self, p1, p2):
        ...
    def add_dist_to_source(self, XX, source_coordinate=[436642.70,3681927.09], col_name='dist_to_source'):
        ...


import numpy as np
import pandas as pd
from tqdm import tqdm
import scipy.stats as stats
from statsmodels.nonparametric.smoothers_lowess import lowess

from pylenm2.stats import metrics
from pylenm2.stats import gp as stats_gp
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

preprocess_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)



# Helper function for plot_correlation
# Sorts analytes in a specific order: 'TRITIUM', 'URANIUM-238','IODINE-129','SPECIFIC CONDUCTANCE', 'PH', 'DEPTH_TO_WATER'
def _custom_analyte_sort(analytes):
    my_order = 'TURISPDABCEFGHJKLMNOQVWXYZ-_abcdefghijklmnopqrstuvwxyz135790 2468'
    return sorted(
        analytes, 
        key=lambda word: [
            my_order.index(c) for c in word
        ],
    )


def get_MCL(analyte_name):
    """Returns the Maximum Concentration Limit value for the specified analyte. Example: 'TRITIUM' returns 1.3

    Args:
        analyte_name (str): name of the analyte to be processed

    Returns:
        float: MLC value
    """
    mcl_dictionary = {
        'TRITIUM': 1.3, 
        'URANIUM-238': 1.31, 
        'NITRATE-NITRITE AS NITROGEN': 1,
        'TECHNETIUM-99': 2.95, 
        'IODINE-129': 0, 
        'STRONTIUM-90': 0.9,
    }
    return mcl_dictionary[analyte_name]


def remove_outliers(
        data, 
        z_threshold=4,
        nan_policy="propagate",     # same as the default of stats.zscore()
    ):
    """Removes outliers from a dataframe based on the z_scores and returns the new dataframe.
    NOTE: The new logic is not same as the previous logic. 
        As per the new logic, it sets NA for the values that are greater than 
        z_threshold, whereas, in the previous logic, all the rows containing 
        even a single column of z > threshold are dropped (and then NaNs are added internally for whatever rows are dropped.)
    TODO: Confirm if this change is okay or if we need to revert back to the previous implementation. @Zexuan
    
    TODO: REWRITE!!! 
        REMOVE `nan_policy`.
        TEST AGAIN AFTER REWRITING TO MAKE SURE THAT THE OUTPUTS MATCH.

    Args:
        data (pd.DataFrame): data for the outliers to removed from
        z_threshold (int, optional): z_score threshold to eliminate. Values above this threshold are elimited. Defaults to 4. 
            NOTE: Get the threshold and logic confirmed by the @Zexuan and @Haruko.
        nan_policy (str, optional): specifies how to handle `nan` values. Passed in the stats.zscore() function.
            Options are one of:
                'propagate': returns nan.
                'raise': throws an error.
                'omit': performs the calculations ignoring nan values.
            Default is 'omit'.

    Returns:
        pd.DataFrame: data with outliers removed
    """

    # Old implementation
    z = np.abs(stats.zscore(data, nan_policy=nan_policy))
    row_loc = np.unique(np.where(z > z_threshold)[0])
    
    # Setting values outside threshold to `nan` values.
    # data = data.drop(data.index[row_loc])
    try:
        data = data.drop(data.index[row_loc]).reindex(data.index)     # NOTE: Reindexing is necessary to make sure that the size mismatch is handled by adding `NaN` values for the dropped rows.
    
    except ValueError as ve:
        preprocess_logger.error(ve)
        data = data.drop(data.index[row_loc])   # NOTE: Using this as a hack currently.
    
    except Exception as e:
        raise(e)

    # # New implementation
    # z = np.abs(stats.zscore(data, nan_policy=nan_policy))
    # data[z > z_threshold] = np.nan

    return data


## by K. Whiteaker 2024-08, kwhit@alum.mit.edu
def remove_outliers_lowess(
        # self, 
        data, 
        lowess_frac=0.1, 
        std_thresh=2.2, 
        return_difference=False,
    ):
    """Identifies outliers in time series data as deviations of std_thresh standard deviations from a LOWESS fit, then sets these outliers to np.nan

    Args:
        data (pd.Series, pd.DataFrame): data for the outliers to removed from. 
            MUST BE indexed by datetime (ex. collection date). 
            Can convert pylenm concentration dataframe (for one station and 
            one analyte) into a valid input via 
            dataframe.set_index('COLLECTION_DATE').RESULT. 
            Expects a single column of data if pd.DataFrame.
        lowess_frac (float, optional): fraction of total data points considered 
            in local fits of LOWESS smoother. A smaller value captures more 
            local behaviour, and may be required for large datasets. 
            Defaults to 0.1.
        std_thresh (int, optional): number of standard deviations in 
            (observation - fit) outside of which is considered an outlier. 
            Defaults to 2.2.
        return_difference (bool, optional): if True, return a pd.Series 
            containing the difference between data and lowess fit

    Returns:
        if return_difference:
            pd.Series: input data with outliers set to np.nan
            pd.Series: input data - lowess fit
        else:
            pd.Series: input data with outliers set to np.nan
    """

    # create a copy so the data isn't modified inplace
    working_data = data.copy()
    
    # lowess() reads datetime as nanoseconds and has issues with large numbers & low frac, so need to scale down x-axis during fitting
    scaleDown = 1e17
    x_data = pd.to_datetime(working_data.index)
    x_readable = x_data.astype(int).to_numpy()/scaleDown
    data_lowess = lowess(
        working_data, 
        x_readable, 
        frac=lowess_frac, 
        return_sorted=False,
    )

    # identify outlier locations and mark them as nan
    difference = working_data - data_lowess  # this is a pd.Series
    thresh = std_thresh * np.std(difference)
    difference_ignoreNaN = np.ma.array(difference, mask=np.isnan(difference)) # Use a mask to mark & ignore the NaNs as per https://stackoverflow.com/questions/37749900/how-to-disregard-the-nan-data-point-in-numpy-array-and-generate-the-normalized-d
    outliers_iloc = np.where(np.abs(difference_ignoreNaN)>thresh)[0]
    outliers = working_data.iloc[outliers_iloc]
    working_data.iloc[outliers_iloc] = np.nan

    if return_difference:
        return working_data, difference
    else:
        return working_data


# Helper fucntion for get_Best_Stations
def _get_Best_Station(
        # self, 
        X, 
        y, 
        xx, 
        ref, 
        selected, 
        leftover, 
        ft=['Elevation'], 
        regression='linear', 
        verbose=True, 
        smooth=True, 
        model=None,
    ):

    num_selected = len(selected)
    errors = []
    
    if(model==None):
        if(len(selected)<5):
            model, pred = stats_gp.fit_gp(X, y, xx)
        else:
            model = None
    else:
        model=model
    
    if(verbose):  
        print("# of stations to choose from: ", len(leftover))
    
    if num_selected == 0:
        if(verbose): 
            print("Selecting first station")
        for ix in leftover:
            y_pred, r_map, residuals, lr_trend = stats_gp.interpolate_topo(X=X.iloc[ix:ix+1,:], y=y[ix:ix+1], xx=xx, ft=ft, regression=regression, model=model, smooth=smooth)
            y_err = stats_gp.mse(ref, y_pred)
            errors.append((ix, y_err))
    
    if num_selected > 0:
        for ix in leftover:
            joined = selected + [ix]
            y_pred, r_map, residuals, lr_trend = stats_gp.interpolate_topo(X=X.iloc[joined,:], y=y[joined], xx=xx, ft=ft, regression=regression, model=model, smooth=smooth)
            y_err = metrics.mse(ref, y_pred)
            errors.append((ix, y_err))
        
    err_ix = [x[0] for x in errors]
    err_vals = [x[1] for x in errors]
    min_val = min(err_vals)
    min_ix = err_ix[err_vals.index(min(err_vals))]
    
    preprocess_logger.info(f"Selected station: {min_ix} with a MSE error of {min_val}.")
    
    if(verbose):
        # print("Selected station: {} with a MSE error of {}\n".format(min_ix, min_val))
        print(f"Selected station: {min_ix} with a MSE error of {min_val}.")
    
    return min_ix, min_val


def get_Best_Stations(
        # self, 
        X, 
        y, 
        xx, 
        ref, 
        initial, 
        max_stations, 
        ft=['Elevation'], 
        regression='linear', 
        verbose=True, 
        smooth=True, 
        model=None,
    ):
    """Greedy optimization function to select a subset of stations as to minimizes the MSE from a reference map

    Args:
        X (numpy.array): array of dimension (number of stations, 2) where each element is a pair of UTM coordinates.
        y (numpy.array): array of size (number of stations) where each value corresponds to a concentration value at a station.
        xx (numpy.array): prediction locations
        ref (numpy.array): reference field to optimize for (aka best/true map)
        initial (list): indices of stations as the starting stations for optimization
        max_stations (int): number of stations to optimize for
        ft (list, optional): feature names to train on. Defaults to ['Elevation'].
        regression (str, optional): choice between 'linear' for linear regression, 'rf' for random forest regression, 'ridge' for ridge regression, or 'lasso' for lasso regression.. Defaults to 'linear'.
        verbose (bool, optional): v. Defaults to True.
        smooth (bool, optional): flag to toggle WhiteKernel on and off. Defaults to True.
        model (GaussianProcessRegressor, optional): model to fit. Defaults to None.

    Returns:
        list: index of best stations in order from best to worst
    """

    BAD_RETURN = None, None

    if not isinstance(initial, list):
        preprocess_logger.error(f"`initial` parameter must be a list. Found `{type(initial)}`.")
        return BAD_RETURN

    tot_err = []
    selected = initial
    leftover = list(range(0, X.shape[0])) # all indexes from 0 to number of station
    
    # Remove the initial set of stations from pool of station indices to choose from
    for i in initial:
        leftover.remove(i)

    station_itr_count = max_stations - len(selected)
    for i in tqdm(range(station_itr_count), desc="Station", total=station_itr_count):

        # For some reason both the if...else conditions contain the same piece of code!!!
        # if i == 0: # select first station will min error
        #     station_ix, err = _get_Best_Station(
        #         X=X, 
        #         y=y, 
        #         xx=xx, 
        #         ref=ref, 
        #         selected=selected, 
        #         leftover=leftover, 
        #         ft=ft, 
        #         regression=regression, 
        #         verbose=verbose, 
        #         smooth=smooth, 
        #         model=model,
        #     )
        #     selected.append(station_ix)
        #     leftover.remove(station_ix)
        #     tot_err.append(err)
        
        # else:
        station_ix, err = _get_Best_Station(
            X=X, 
            y=y, 
            xx=xx, 
            ref=ref, 
            selected=selected, 
            leftover=leftover, 
            ft=ft, 
            regression=regression, 
            verbose=verbose, 
            smooth=smooth, 
            model=model,
        )
        selected.append(station_ix)
        leftover.remove(station_ix)
        tot_err.append(err)
    
    preprocess_logger.info(f"{selected = }")
    
    return selected, tot_err



import os
import scipy
import random
import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.dates import date2num, num2date

from supersmoother import SuperSmoother
from statsmodels.nonparametric.smoothers_lowess import lowess

from pylenm2.data import filters
from pylenm2.stats import preprocess
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

plots_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


# def _plotUpperHalf(*args, **kwargs):
def _plotUpperHalf(x_data, y_data, *args, **kwargs):   # TODO: confirm that this change does not break any other function.
    """<Function docstring>
    TODO: Complete the function docstring.

    Args:
        x_data (pd.Series): X-axis data.
        y_data (pd.Series): Y-axis data.
        args (Tuple): Tuple of positional arguments.
        kwargs (Dict): Dictionary of keyword arguments.
    """

    # corr_r = args[0].corr(args[1], 'pearson')
    corr_r = x_data.corr(y_data, 'pearson')
    corr_text = f"{corr_r:2.2f}"
    
    ax = plt.gca()
    ax.set_axis_off()
    
    marker_size = abs(corr_r) * 10000
    ax.scatter([.5], [.5], marker_size, [corr_r], alpha=0.6, cmap="coolwarm",
                vmin=-1, vmax=1, transform=ax.transAxes)
    
    font_size = abs(corr_r) * 40 + 5
    ax.annotate(corr_text, [.5, .48,],  xycoords="axes fraction", # [.5, .48,]
                ha='center', va='center', fontsize=font_size, fontweight='bold')


def plot_data(
        data_pylenm_dm, 
        station_name, 
        analyte_name, 
        log_transform=True, 
        alpha=0,
        plot_inline=True, 
        year_interval=2, 
        x_label='Years', 
        y_label='',
        save_plot=False, 
        save_dir='plot_data', 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Plot concentrations over time of a specified station and analyte with a smoothed curve on interpolated data points.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analyte_name (str): name of the analyte to be processed
        log_transform (bool, optional): choose whether or not the data should 
            be transformed to log base 10 values. Defaults to `True`.
        alpha (int, optional): value between 0 and 10 for line smoothing. 
            Defaults to 0.
        plot_inline (bool, optional): `True` to show plot inline else `False`. 
            Defaults to `True`.
        year_interval (int, optional): plot by how many years to appear in the 
            axis e.g.(1 = every year, 5 = every 5 years, ...). Defaults to 2.
        x_label (str, optional): x axis label. Defaults to 'Years'.
        y_label (str, optional): y axis label. Defaults to ''.
        save_plot (bool, optional): `True` to save the plot to the `save_dir` 
            else `False`. Defaults to `False`.
        save_dir (str, optional): name of the directory you want to save the 
            plot to. `save_plot` must be `True` for this to be useful. 
            Defaults to 'plot_data'.
        filter (bool, optional): flag to indicate filtering. 
            Defaults to `False`.
        col (str, optional): column to filter. Example: col='STATION_ID'. 
            Defaults to None.
        equals (list, optional): values to filter col by. 
            Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].

    Returns:
        None
    """

    # Gets appropriate data (station_name and analyte_name)
    query_raw = filters.query_data(
        data_pylenm_dm=data_pylenm_dm, 
        station_name=station_name, 
        analyte_name=analyte_name,
    )       # TODO: Handle better for empty values.

    query = filters.simplify_data(data=query_raw)
    
    # Check if the query returned any results
    # if(type(query)==int and query == 0):
    if query is None:
        plots_logger.error(f"No results found for {station_name} and {analyte_name}")
        return None
    
    else:
        # Filter the data by the specified column and values (if necessary)
        if(filter):
            filter_res = filters.filter_by_column(
                data=data_pylenm_dm.construction_data, 
                col=col, 
                equals=equals,
            )
            
            if('ERROR:' in str(filter_res)):
                plots_logger.error("Ran into ERROR when calling filter_by_column()!")
                return filter_res

            # Get the intersection of the query stations and filter stations
            query_stations = list(query.STATION_ID.unique())
            filter_stations = list(filter_res.index.unique())
            intersect_stations = list(set(query_stations) & set(filter_stations))
            if(len(intersect_stations)<=0):
                plots_logger.error('ERROR: No results for this query with the specifed filter parameters.')
                return 'ERROR: No results for this query with the specifed filter parameters.'
            query = query[query['STATION_ID'].isin(intersect_stations)]
        
        # Extract the date and result values from the query data
        x_data = query.COLLECTION_DATE
        x_data = pd.to_datetime(x_data)
        y_data = query.RESULT
        if(log_transform):
            y_data = np.log10(y_data)

        # Prepare data for Supersmoother
        x_RR = x_data.astype(int).to_numpy()
        nu = x_data.shape[0]

        # Create and fit supersmoother model
        model = SuperSmoother(alpha=alpha)
        try:
            model.fit(x_RR, y_data)

            # Get smoothed predictions, residuals and outliers from the model.
            y_pred = model.predict(x_RR)
        
        except Exception as e:
            plots_logger.error(e)
            return None

        r = model.cv_residuals()
        out = abs(r) > (2.2 * np.std(r))
        out_x = x_data[out]
            
        out_y = y_data[out]

        # Plot data
        plt.figure(figsize=(8,8))
        ax = plt.axes()
        years = mdates.YearLocator(year_interval)  # every year
        months = mdates.MonthLocator()  # every month
        yearsFmt = mdates.DateFormatter('%Y')

        for label in ax.get_xticklabels():
            label.set_rotation(30)
            label.set_horizontalalignment('center')

        ax = plt.gca()
        ax.xaxis.set_major_locator(years)
        ax.xaxis.set_major_formatter(yearsFmt)
        ax.autoscale_view()

        unit = query.RESULT_UNITS.values[0]

        ax.set_title(
            # str(station_name) + ' - ' + analyte_name, 
            f"{station_name} - {analyte_name}", fontweight='bold',
        )
        ttl = ax.title
        ttl.set_position([.5, 1.05])
        
        if(y_label==''):    
            if(log_transform):
                ax.set_ylabel('log-Concentration (' + unit + ')')
            else:
                ax.set_ylabel('Concentration (' + unit + ')')
        
        else:
            ax.set_ylabel(y_label)
        
        ax.set_xlabel(x_label)
        small_fontSize = 15
        large_fontSize = 20
        
        plt.rc('axes', titlesize=large_fontSize)
        plt.rc('axes', labelsize=large_fontSize)
        plt.rc('legend', fontsize=small_fontSize)
        plt.rc('xtick', labelsize=small_fontSize)
        plt.rc('ytick', labelsize=small_fontSize) 
        
        ax.plot(x_data, y_data, ls='', marker='o', ms=5, color='black', alpha=1)
        ax.plot(x_data, y_pred, ls='-', marker='', ms=5, lw=2, color='blue', alpha=0.5, label="Super Smoother")
        ax.plot(out_x , out_y, ls='', marker='o', ms=5, color='red', alpha=1, label="Outliers")
        ax.legend(bbox_to_anchor=(1.04, 1), loc='upper left', borderaxespad=0.)
        
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)       
        ax.text(1.05, 0.85, 'Samples: {}'.format(nu), transform=ax.transAxes, 
                fontsize=small_fontSize,
                fontweight='bold',
                verticalalignment='top', 
                bbox=props)
        
        # Save the plot to `save_dir` (if necessary)
        if save_plot:
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            plt.savefig(
                f"{save_dir}/{str(station_name)}-{analyte_name}.png", 
                bbox_inches="tight",
            )
        
        # Show plot inline (used in notebooks)
        if(plot_inline):
            plt.show()
        
        # Clear all the plots from memory
        plt.clf()
        plt.cla()
        plt.close()

        # Returning a non-None value to mark a successful execution
        return 200


def plot_all_data(
        # self, 
        data_pylenm_dm,
        log_transform=True, 
        alpha=0, 
        year_interval=2, 
        plot_inline=True, 
        save_dir='plot_data',
    ):
    """Plot concentrations over time for every station and analyte with a smoothed curve on interpolated data points.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        log_transform (bool, optional): choose whether or not the data should be transformed to log base 10 values. Defaults to True.
        alpha (int, optional): alue between 0 and 10 for line smoothing. Defaults to 0.
        plot_inline (bool, optional): choose whether or not to show plot inline. Defaults to True.
        year_interval (int, optional): plot by how many years to appear in the axis e.g.(1 = every year, 5 = every 5 years, ...). Defaults to 2.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_data'.
    """
    
    analytes = ['TRITIUM','URANIUM-238','IODINE-129','SPECIFIC CONDUCTANCE', 'PH', 'DEPTH_TO_WATER']
    
    stations = np.array(data_pylenm_dm.data.STATION_ID.values)
    stations = np.unique(stations)

    stations_analytes_iter = list(itertools.product(stations, analytes))
    
    success = 0
    errors = 0
    
    # for station in stations:
    #     for analyte in analytes:
    for station, analyte in tqdm(stations_analytes_iter, desc="Station Analyte Pair", total=(len(analytes)*len(stations))):
        plot = plot_data(
            data_pylenm_dm=data_pylenm_dm,
            station_name=station, 
            analyte_name=analyte, 
            log_transform=log_transform, 
            alpha=alpha, 
            year_interval=year_interval, 
            plot_inline=plot_inline, 
            save_plot=True,
            save_dir=save_dir,
        )
        
        if plot is None:
            errors = errors + 1
        else:
            success = success + 1

    plots_logger.info(f"Success: {success}")
    plots_logger.info(f"Errors: {errors}")

        
def plot_MCL(
        # self, 
        data_pylenm_dm,
        station_name, 
        analyte_name, 
        year_interval=5, 
        save_dir='plot_MCL',
    ):
    """Plots the linear regression line of data given the analyte_name and station_name. The plot includes the prediction where the line of best fit intersects with the Maximum Concentration Limit (MCL).

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): ame of the station to be processed
        analyte_name (str): name of the analyte to be processed
        year_interval (int, optional): lot by how many years to appear in the axis e.g.(1 = every year, 5 = every 5 years, ...). Defaults to 5.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_MCL'.
    """
    
    # data = self.data
    data = data_pylenm_dm.data
    
    # finds the intersection point of 2 lines given the slopes and y-intercepts
    def line_intersect(m1, b1, m2, b2):
        if m1 == m2:
            print ('The lines are parallel')
            return None
        x = (b2 - b1) / (m1 - m2)
        y = m1 * x + b1
        return x,y

    # Gets appropriate data (station_name and analyte_name)
    # query = self.query_data(station_name, analyte_name)
    query = filters.query_data(
        data_pylenm_dm=data_pylenm_dm,
        station_name=station_name, 
        analyte_name=analyte_name,
    )

    # if(type(query)==int and query == 0):
    if isinstance(query, int) and query==0:
        plots_logger.info(
            f"No results found for {station_name} and {analyte_name}"
        )
        return None
    
    else:
        test = query.groupby(['COLLECTION_DATE'])['RESULT'].mean()
        test.index = pd.to_datetime(test.index)

        x = date2num(test.index)
        # y = np.log10(test.RESULT)
        y = np.log10(test)
        ylabel = 'log-Concentration (' + data_pylenm_dm.get_unit(analyte_name) + ')'
        y = y.rename(ylabel)

        p, cov = np.polyfit(x, y, 1, cov=True)  # parameters and covariance from of the fit of 1-D polynom.

        m_unc = np.sqrt(cov[0][0])
        b_unc = np.sqrt(cov[1][1])

        f = np.poly1d(p)

        try:
            # MCL = self.get_MCL(analyte_name)
            MCL = preprocess.get_MCL(analyte_name)
            m1, b1 = f # line of best fit
            m2, b2 = 0, MCL # MCL constant

            intersection = line_intersect(m1, b1, m2, b2)

            ## Get confidence interval intersection points with MCL
            data = list(zip(x,y))
            n = len(data)
            list_slopes = []
            list_intercepts = []
            random.seed(50)
            
            for _ in range(80):
                sampled_data = [ random.choice(data) for _ in range(n) ]
                x_s, y_s = zip(*sampled_data)
                x_s = np.array(x_s)
                y_s = np.array(y_s)

                m_s, b_s, r, p, err = scipy.stats.linregress(x_s,y_s)
                ymodel = (m_s * x_s) + b_s
                
                list_slopes.append(m_s)
                list_intercepts.append(b_s)

            max_index = list_slopes.index(max(list_slopes))
            min_index = list_slopes.index(min(list_slopes))
            
            intersection_left = line_intersect(list_slopes[min_index], list_intercepts[min_index], m2, b2)
            intersection_right = line_intersect(list_slopes[max_index], list_intercepts[max_index], m2, b2)
            ##

            fig, ax = plt.subplots(figsize=(10, 6))

            ax.set_title(station_name + ' - ' + analyte_name, fontweight='bold')
            ttl = ax.title
            ttl.set_position([.5, 1.05])
            years = mdates.YearLocator(year_interval)  # every year
            months = mdates.MonthLocator()  # every month
            yearsFmt = mdates.DateFormatter('%Y') 

            for label in ax.get_xticklabels():
                label.set_rotation(30)
                label.set_horizontalalignment('center')

            ax.xaxis.set_major_locator(years)
            ax = plt.gca()
            ax.xaxis.set_major_locator(years)
            ax.xaxis.set_major_formatter(yearsFmt)
            ax.autoscale_view()
            ax.grid(True, alpha=0.4)
            small_fontSize = 15
            large_fontSize = 20
            
            plt.rc('axes', titlesize=large_fontSize)
            plt.rc('axes', labelsize=large_fontSize)
            plt.rc('legend', fontsize=small_fontSize)
            plt.rc('xtick', labelsize=small_fontSize)
            plt.rc('ytick', labelsize=small_fontSize)

            ax.set_xlabel('Years')
            ax.set_ylabel('log-Concentration (' + data_pylenm_dm.get_unit(analyte_name) + ')')

            if(intersection[0] < min(x)):
                temp = intersection_left
                intersection_left = intersection_right
                intersection_right = temp
                ax.set_ylim([0, max(y)+1])
                ax.set_xlim([intersection_left[0]-1000, max(x)+1000])
            
            elif(intersection[0] < max(x) and intersection[0] > min(x)):
                ax.set_ylim([0, max(y)+1])
                ax.set_xlim(min(x)-1000, max(x)+1000)

            else:
                ax.set_ylim([0, max(y)+1])
                ax.set_xlim([min(x)-1000, intersection_right[0]+1000])

            ax = sns.regplot(
                x=x, y=y, logx=True, truncate=False, 
                seed=42, n_boot=1000, ci=95,
            ) # Line of best fit
            ax.plot(x, y, ls='', marker='o', ms=5, color='black', alpha=1) # Data
            ax.axhline(y=MCL, color='r', linestyle='--') # MCL
            ax.plot(intersection[0], intersection[1], color='blue', marker='o', ms=10)
            ax.plot(intersection_left[0], intersection_left[1], color='green', marker='o', ms=5)
            ax.plot(intersection_right[0], intersection_right[1], color='green', marker='o', ms=5)

            predict = num2date(intersection[0]).date()
            l_predict = num2date(intersection_left[0]).date()
            u_predict = num2date(intersection_right[0]).date()
            
            ax.annotate(
                predict, 
                (intersection[0], intersection[1]), 
                xytext=(intersection[0], intersection[1]+1), 
                bbox=dict(boxstyle="round", alpha=0.1), 
                ha='center', 
                arrowprops=dict(arrowstyle="->", color='blue'), 
                fontsize=small_fontSize, 
                fontweight='bold',
            )
            
            props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
            ax.text(
                1.1, 
                0.5, 
                # 'Lower confidence: {}\nPrediction: {}\nUpper confidence: {}'.format(l_predict, predict, u_predict), 
                f"Lower confidence: {l_predict}\nPrediction: {predict}\nUpper confidence: {u_predict}", 
                transform=ax.transAxes, 
                fontsize=small_fontSize, 
                fontweight='bold', 
                verticalalignment='bottom', 
                bbox=props,
            )

            # Save figure
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            
            plt.savefig(
                # save_dir + '/' + station_name + '-' + analyte_name +'.png', 
                f"{save_dir}/{station_name}-{analyte_name}.png", 
                bbox_inches="tight",
            )

        except Exception as e:
            plots_logger.error('ERROR: Something went wrong')
            plots_logger.error(e)
            return None



## by K. Whiteaker 2024-08, kwhit@alum.mit.edu
# Helper function for plot_data_weekAvg() and plot_data_lowess()
# copied from self.plot_data(), with a few additional plot items added
def _plot_data_xOutliers_fit(
        # self, 
        station_name, 
        analyte_name, 
        concentration_data_xOutliers, 
        concentration_data_xOutliers_fit, 
        outliers, 
        units,
        difference, 
        fitColor, 
        fitName, 
        rm_outliers, 
        std_thresh, 
        lowess_frac, 
        show_difference, 
        x_label, 
        y_label, 
        year_interval, 
        log_transform, 
        y_zoom, 
        return_data, 
        save, 
        save_dir, 
        plot_inline, 
        save_as_pdf,
    ):
    
    # define figure shape and size
    plt.figure(figsize=(8,8))
    ax = plt.axes()
    
    years = mdates.YearLocator(year_interval)  # every year
    months = mdates.MonthLocator()  # every month
    yearsFmt = mdates.DateFormatter('%Y')
    
    for label in ax.get_xticklabels():
        label.set_rotation(30)
        label.set_horizontalalignment('center')
    
    ax = plt.gca()
    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_formatter(yearsFmt)
    ax.autoscale_view()
    
    if y_label is None:
        y_label = analyte_name + f' [{units}]'
        if(log_transform):
            y_label = 'log_10 of ' + analyte_name + f' [{units}]'
    
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    
    small_fontSize = 15
    large_fontSize = 20
    plt.rc('axes', titlesize=large_fontSize)
    plt.rc('axes', labelsize=large_fontSize)
    plt.rc('legend', fontsize=small_fontSize)
    plt.rc('xtick', labelsize=small_fontSize)
    plt.rc('ytick', labelsize=small_fontSize)
    
    props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
    ax.text(
        1.05, 
        0.85, 
        f"Samples: {concentration_data_xOutliers.size}", 
        transform=ax.transAxes, 
        fontsize=small_fontSize,
        fontweight='bold',
        verticalalignment='top', 
        bbox=props,
    )
    
    # ax.set_title(str(station_name) + ' - ' + analyte_name, fontweight='bold')
    ax.set_title(f"{station_name} - {analyte_name}", fontweight='bold')
    ttl = ax.title
    ttl.set_position([.5, 1.05])

    # plot items
    ax.plot(
        concentration_data_xOutliers.index, 
        concentration_data_xOutliers, 
        ls='', 
        marker='o', 
        ms=2, 
        color='black', 
        alpha=1,
    )
    
    if rm_outliers:
        ax.plot(
            outliers.index, 
            outliers, 
            ls='', 
            marker='x', 
            ms=5, 
            color='red', 
            alpha=0.75, 
            label=f"Outliers ({outliers.size})",
        )
    
    # need to use pd.Series.dropna() when plotting so that matplotlib doesn't insert any gaps
    ax.plot(
        concentration_data_xOutliers_fit.dropna().index, 
        concentration_data_xOutliers_fit.dropna(), 
        ls='-', 
        marker='', 
        ms=2, 
        lw=1, 
        color=fitColor, 
        alpha=1, 
        label=fitName,
    )
    
    if y_zoom:
        plt.ylim(min(concentration_data_xOutliers), max(concentration_data_xOutliers))
    ax.legend(bbox_to_anchor=(1.04, 1), loc='upper left', borderaxespad=0.)
    
    fileType = '.pdf' if save_as_pdf else '.png'
    
    if save:
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        plt.savefig(save_dir + '/' + str(station_name) + '_' + analyte_name + '_' + fitName.replace(" ", "") + fileType, bbox_inches="tight")
    
    if(plot_inline):
        plt.show()
    
    plt.clf()
    plt.cla()
    plt.close()
    
    # plot difference between fit and observation, with outliers marked
    if rm_outliers and show_difference:
        outliers_loc = outliers.index
        
        plt.scatter(difference.index, difference, s=0.5, label='Difference')
        
        plt.scatter(difference[outliers_loc].index, difference[outliers_loc], s=10, marker='x', label=f'Outliers outside {std_thresh} SD')
        
        plt.title("(obs - fit), fraction "+ str(lowess_frac) + ", station "+ str(station_name))
        
        plt.ylabel('Observation - LOWESS Fit [ft]')
        plt.legend()
        
        if save:
            # plt.savefig(save_dir + '/' + str(station_name) + '_' + analyte_name + '_difference' + fileType, bbox_inches="tight")
            plt.savefig(
                f"{save_dir}/{station_name}_{analyte_name}_difference{fileType}", 
                bbox_inches="tight",
            )
        
        if(plot_inline):
            plt.show()
        
        plt.clf()
        plt.cla()
        plt.close()


## by K. Whiteaker 2024-08, kwhit@alum.mit.edu
def plot_data_rollAvg(
        # self, 
        data_pylenm_dm,
        station_name, 
        analyte_name, 
        window='1W', 
        rm_outliers=True, 
        std_thresh=2.2, 
        lowess_frac=0.1, 
        show_difference=False, 
        x_label=None, 
        y_label=None, 
        year_interval=2, 
        log_transform=False, 
        y_zoom=False, 
        return_data=False, 
        save=False, 
        save_dir='plot_data_lowess',
        plot_inline=True, 
        save_as_pdf=False,
    ):
    """Plot time series data for a specified station and analyte, alongside a curve calculated via a rolling time average (default 1 week) centred around each point

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analyte_name (str): name of the analyte to be processed
        window (str): the time length of the rolling window used for averaging, ex. '1W', '2D'. Defaults to '1W'. Months are not supported; valid inputs are listed here: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html
        rm_outliers (bool, optional): if True, use lowess function to remove outliers. Defaults to True
        std_thresh (int, optional): number of standard deviations in (observation - LOWESS fit) outside of which is considered an outlier. Defaults to 2.2
        lowess_frac (float, optional): fraction of total data points considered in local fits of LOWESS smoother. A smaller value captures more local behaviour, and may be required for large datasets. Defaults to 0.1
        show_difference (bool, optional): if True, plots (observation - LOWESS fit) at each measurement time
        x_label (str, optional): x axis label. Defaults to None
        y_label (str, optional): y axis label. Defaults to None. If left as None, will be set to analyte_name + f' [{units}]'
        year_interval (int, optional): plot by how many years to appear in the axis e.g.(1 = every year, 5 = every 5 years, ...). Defaults to 2
        log_transform (bool, optional): choose whether or not the data should be transformed to log base 10 values. Defaults to False
        y_zoom (bool, optional): if True, plot y axes will zoom in to [minimum y value, maximum y value] after removing outliers. Defaults to False
        return_data (bool, optional): if True, return the data used to make the plots. Defaults to False
        save (bool, optional): if True, save plots to file in save_dir. Defaults to False
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_data_lowess'
        plot_inline (bool, optional): choose whether or not to show plot inline. Defaults to True
        save_as_pdf (bool, optional): if True, saves figure as vectorized pdf instead of png. Defaults to False

    Returns:
        if return_data:
            if rm_outliers:
                pd.Series: concentration data with outliers set to np.nan, indexed by date
                pd.Series: rolling average of post-outlier-removal concentration data, indexed by date
                pd.Series: outliers, indexed by date
            else:
                pd.Series: concentration data, indexed by date
                pd.Series: rolling average of concentration data, indexed by date
        else:
            None
    """

    # import data and filter for the chosen analyte
    query = filters.query_data(
        data_pylenm_dm=data_pylenm_dm,
        station_name=station_name, 
        analyte_name=analyte_name,
    )
    query = filters.simplify_data(data=query)
    
    # if(type(query)==int and query == 0):
    if query is None:
        # return 'No results found for {} and {}'.format(station_name, analyte_name)
        plots_logger.error(f"No results found for {station_name} and {analyte_name}.")
        return None
    
    # reshape data so it can be passed into lowess() and remove_outliers_lowess()
    concentration_data = query.set_index('COLLECTION_DATE').RESULT
    time_data = concentration_data.index
    units = query.RESULT_UNITS.values[0]  # assumes same units for every entry
    
    if(log_transform):
        concentration_data = concentration_data.apply(np.log10)

    # if rm_outliers toggled, call remove_outliers_lowess() to set outliers to np.nan via lowess fit
    if rm_outliers:
        concentration_data_xOutliers, difference = preprocess.remove_outliers_lowess(
            concentration_data, 
            lowess_frac=lowess_frac,
            std_thresh=std_thresh, 
            return_difference=True,
        )
    else:
        concentration_data_xOutliers = concentration_data
        difference = pd.Series  # will not be used, just need something to pass into _plot_data_xOutliers_fit
    
    outliers_loc = concentration_data_xOutliers.compare(concentration_data).index  # dates at which outliers occurred
    outliers = concentration_data[outliers_loc]  # outlier values indexed by their date of occurrence
    num_outliers = outliers.size
    analyte_max_xOutliers, analyte_min_xOutliers = max(concentration_data_xOutliers), min(concentration_data_xOutliers)
            
    # now that outliers have been dropped, calculate rolling average of y values (result of average is assigned to the centre of the window)
    avg_window = pd.Timedelta(window)  # could also do pd.Timedelta(7, "d")
    rolling_ser = concentration_data_xOutliers
    roll = rolling_ser.rolling(
        window=avg_window, 
        min_periods=1, 
        center=True,
    )
    concentration_data_xOutliers_avg = roll.mean()
    
    # plot rolling average
    _plot_data_xOutliers_fit(
        station_name=station_name, 
        analyte_name=analyte_name, 
        concentration_data_xOutliers=concentration_data_xOutliers, 
        concentration_data_xOutliers_fit=concentration_data_xOutliers_avg, 
        outliers=outliers, 
        units=units, 
        difference=difference, 
        fitColor='green', 
        fitName=f"{window} Average", 
        rm_outliers=rm_outliers, 
        std_thresh=std_thresh, 
        lowess_frac=lowess_frac, 
        show_difference=show_difference, 
        x_label=x_label, 
        y_label=y_label, 
        year_interval=year_interval, 
        log_transform=log_transform, 
        y_zoom=y_zoom, 
        return_data=return_data, 
        save=save, 
        save_dir=save_dir, 
        plot_inline=plot_inline, 
        save_as_pdf=save_as_pdf,
    )
    
    # return data
    if return_data:
        if rm_outliers:
            return concentration_data_xOutliers, concentration_data_xOutliers_avg, outliers
        else:
            return concentration_data_xOutliers, concentration_data_xOutliers_avg
        

## by K. Whiteaker 2024-08, kwhit@alum.mit.edu
def plot_data_lowess(
        # self, 
        data_pylenm_dm, 
        station_name, 
        analyte_name, 
        rm_outliers=True, 
        std_thresh=2.2, 
        lowess_frac=0.1, 
        show_difference=False, 
        x_label=None, 
        y_label=None, 
        year_interval=2, 
        log_transform=False, 
        y_zoom=False, 
        return_data=False, 
        save=False, 
        save_dir='plot_data_lowess', 
        plot_inline=True, 
        save_as_pdf=False,
    ):
    """Plot time series data for a specified station and analyte, alongside a 
    smoothed curve on interpolated data points (LOWESS).
    
    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object 
            containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analyte_name (str): name of the analyte to be processed
        rm_outliers (bool, optional): if True, use lowess function to 
            remove outliers. Defaults to True
        std_thresh (int, optional): number of standard deviations in 
            (observation - fit) outside of which is considered an outlier. 
            Defaults to 2.2
        lowess_frac (float, optional): fraction of total data points 
            considered in local fits of LOWESS smoother. A smaller value 
            captures more local behaviour, and may be required for large 
            datasets. Defaults to 0.1
        show_difference (bool, optional): if True, plots (observation - 
            LOWESS fit) at each measurement time
        x_label (str, optional): x axis label. Defaults to None
        y_label (str, optional): y axis label. Defaults to None. If left as 
            None, will be set to analyte_name + f' [{units}]'
        year_interval (int, optional): plot by how many years to appear in 
            the axis e.g.(1 = every year, 5 = every 5 years, ...). 
            Defaults to 2.
        log_transform (bool, optional): choose whether or not the data 
            should be transformed to log base 10 values. Defaults to False
        y_zoom (bool, optional): if True, plot y axes will zoom in to 
            [minimum y value, maximum y value] after removing outliers. 
            Defaults to False.
        return_data (bool, optional): if True, return the data used to make 
            the plots. Defaults to False.
        save (bool, optional): if True, save plots to file in save_dir. 
            Defaults to False.
        save_dir (str, optional): name of the directory you want to save 
            the plot to. Defaults to 'plot_data_lowess'.
        plot_inline (bool, optional): choose whether or not to show plot 
            inline. Defaults to True.
        save_as_pdf (bool, optional): if True, saves figure as vectorized 
            pdf instead of png. Defaults to False.

    Returns:
        if return_data:
            if rm_outliers:
                pd.Series: concentration data with outliers set to np.nan, indexed by date
                pd.Series: LOWESS fit of post-outlier-removal concentration data, indexed by date
                pd.Series: outliers, indexed by date
            else:
                pd.Series: concentration data, indexed by date
                pd.Series: LOWESS fit of concentration data, indexed by date
        else:
            None
    """
    
    # import data and filter for the chosen analyte
    query = filters.query_data(
        data_pylenm_dm=data_pylenm_dm, 
        station_name=station_name, 
        analyte_name=analyte_name,
    )
    query = filters.simplify_data(data=query)
    
    # if(type(query)==int and query == 0):
    if query is None:
        # return 'No results found for {} and {}'.format(station_name, analyte_name)
        plots_logger.error(f"No results found for {station_name} and {analyte_name}.")
        return None
    
    # reshape data so it can be passed into lowess() and remove_outliers_lowess()
    concentration_data = query.set_index('COLLECTION_DATE').RESULT
    time_data = concentration_data.index
    units = query.RESULT_UNITS.values[0]  # assumes same units for every entry
    
    if(log_transform):
        concentration_data = concentration_data.apply(np.log10)

    # if rm_outliers toggled, call remove_outliers_lowess() to set outliers to np.nan via lowess fit
    if rm_outliers:
        concentration_data_xOutliers, difference = preprocess.remove_outliers_lowess(
            concentration_data, 
            lowess_frac=lowess_frac, 
            std_thresh=std_thresh, 
            return_difference=True,
        )
    else:
        concentration_data_xOutliers = concentration_data
        difference = pd.Series  # will not be used, just need something to pass into __plot_data_xOutliers_fit
    
    outliers_loc = concentration_data_xOutliers.compare(concentration_data).index  # dates at which outliers occurred
    outliers = concentration_data[outliers_loc]  # outlier values indexed by their date of occurrence
    num_outliers = outliers.size
    analyte_max_xOutliers, analyte_min_xOutliers = max(concentration_data_xOutliers), min(concentration_data_xOutliers)
            
    # now that outliers have been dropped, calculate a new lowess fit for plotting
    # lowess() reads datetime as nanoseconds and has issues with large numbers & low frac, so need to scale down x-axis during fitting
    scaleDown = 1e17
    x_readable = time_data.to_numpy().astype(int)/scaleDown
    y_data_lowess = lowess(
        concentration_data_xOutliers, 
        x_readable, 
        frac=lowess_frac, 
        return_sorted=False,
    )
    concentration_data_xOutliers_lowess = pd.Series(
        data=y_data_lowess, 
        index=concentration_data_xOutliers.index, 
        dtype=float, 
        name=concentration_data_xOutliers.name,
    )

    # plot lowess fit
    _plot_data_xOutliers_fit(
        station_name=station_name, 
        analyte_name=analyte_name, 
        concentration_data_xOutliers=concentration_data_xOutliers, 
        concentration_data_xOutliers_fit=concentration_data_xOutliers_lowess, 
        outliers=outliers, 
        units=units, 
        difference=difference, 
        fitColor='violet', 
        fitName='LOWESS Fit', 
        rm_outliers=rm_outliers, 
        std_thresh=std_thresh,
        lowess_frac=lowess_frac, 
        show_difference=show_difference, 
        x_label=x_label, 
        y_label=y_label,
        year_interval=year_interval, 
        log_transform=log_transform, 
        y_zoom=y_zoom, 
        return_data=return_data,
        save=save, 
        save_dir=save_dir, 
        plot_inline=plot_inline, 
        save_as_pdf=save_as_pdf,
    )
    
    if return_data:
        if rm_outliers:
            return concentration_data_xOutliers, concentration_data_xOutliers_lowess, outliers
        else:
            return concentration_data_xOutliers, concentration_data_xOutliers_lowess



import os
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

import pylenm2
from pylenm2.data import filters
from pylenm2.data import fetchers
from pylenm2.stats import preprocess
from pylenm2.data import transformation
from pylenm2.visualization import plots
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

plot_corr_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def _format_ticks_and_labels(axes, fontsize):
    """Utility function to format all the ticks and labels along the `x` and `y` axes.

    Args:
        axes (matplotlib.Axes): List of Axes to format.
        fontsize (int): Fontsize for all the labels and ticks.
    """
    # for ax in g.axes.flat:
    for ax in axes.flat:
        ax.tick_params("y", labelrotation=0, labelsize=fontsize)
        
        # ax.set_xticklabels(
        #     ax.get_xticklabels(), rotation=45, fontsize=fontsize,
        # )   # NOTE: DISCOURAGED version of `xticks` - https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_xticklabels.html
        # ax.set_xticks(
        #     ax.get_xticks(),
        #     [f"{num:.2f}" for num in ax.get_xticks()],    # TODO: Check with Zexuan if this is supposed to be normalized.
        #     rotation=45, 
        #     fontsize=fontsize,
        # )     # NOTE: Non-discouraged version of `xticks`
        ax.tick_params("x", labelrotation=45, labelsize=fontsize)   # NOTE: Cleaner version of `xticks`. TODO: Check if this one works or if we need to revert back to the previous ticklabels. @Zexuan
        ax.set_xlabel(
            ax.get_xlabel(), fontsize=fontsize, fontweight='bold',
        )
        ax.set_ylabel(
            ax.get_ylabel(), fontsize=fontsize, fontweight='bold',
        )


def plot_corr_by_station(
        data_pylenm_dm, 
        station_name, 
        analytes, 
        remove_outliers=True, 
        z_threshold=4, 
        interpolate=False, 
        frequency='2W', 
        save_dir='plot_correlation', 
        log_transform=False, 
        fontsize=20, 
        # returnData=False, 
        remove=[], 
        no_log=None,
    ):
    """Plots the correlations with the physical plots as station as the correlations of the important analytes over time for a specified station.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analytes (list): list of analyte names to use
        remove_outliers (bool, optional): choose whether or to remove the outliers. Defaults to True.
        z_threshold (int, optional): z_score threshold to eliminate outliers. Defaults to 4.
        interpolate (bool, optional): choose whether or to interpolate the data. Defaults to False.
        frequency (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} frequency to interpolate. Note: See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘W’ = every week, ‘D ’= every day, ‘2W’ = every 2 weeks). Defaults to '2W'.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_correlation'.
        log_transform (bool, optional): flag for log base 10 transformation. Defaults to False.
        fontsize (int, optional): font size. Defaults to 20.
        returnData (bool, optional): flag to return data used to perfrom correlation analysis. Defaults to False.
        remove (list, optional): stations to remove. Defaults to [].
        no_log (list, optional): list of column names to not apply log transformation to. Defaults to None.

    Returns:
        None
    """
    
    # Prepare data
    data = data_pylenm_dm.data
    query = data[data.STATION_ID == station_name]
    a = list(np.unique(query.ANALYTE_NAME.values))# get all analytes from dataset
    
    for value in analytes:
        # if((value in a)==False):
        if value not in a:
            plot_corr_logger.error(f"ERROR: No analyte named {value} in data.")
            # return 'ERROR: No analyte named "{}" in data.'.format(value)
            return None

    analytes = sorted(analytes)
    
    query = query.loc[query.ANALYTE_NAME.isin(analytes)]
    x = query[['COLLECTION_DATE', 'ANALYTE_NAME']]
    unique = ~x.duplicated()
    query = query[unique]
    piv = query.reset_index().pivot(index='COLLECTION_DATE',columns='ANALYTE_NAME', values='RESULT')
    piv = piv[analytes]
    piv.index = pd.to_datetime(piv.index)
    totalSamples = piv.shape[0]
    piv = piv.dropna()
    
    # Interpolate (if specified) and define file extension and titles
    if(interpolate):
        piv = transformation.interpolate_station_data(
            data_pylenm_dm=data_pylenm_dm,
            station_name=station_name, 
            analytes=analytes, 
            frequency=frequency,
        )
        # file_extension = '_interpolated_' + frequency
        file_extension = f"_interpolated_{frequency}"
        # title = station_name + '_correlation - interpolated every ' + frequency
        title = f"{station_name}_correlation - interpolated every {frequency}"
    else:
        file_extension = "_correlation"
        # title = station_name + '_correlation'
        title = f"{station_name}_correlation"
    samples = piv.shape[0]
    
    # Plot if there are enough samples
    if(samples < 5):
        if(interpolate):
            # return 'ERROR: {} does not have enough samples to plot.\n Try a different interpolation frequency'.format(station_name)
            plot_corr_logger.error(f"ERROR: {station_name} does not have enough samples to plot.\n Try a different interpolation frequency")
            return None
        
        # return 'ERROR: {} does not have enough samples to plot.'.format(station_name)
        plot_corr_logger.error(f"ERROR: {station_name} does not have enough samples to plot.")
        return None

    else:
        # scaler = StandardScaler()
        # pivScaled = scaler.fit_transform(piv)
        # pivScaled = pd.DataFrame(pivScaled, columns=piv.columns)
        # pivScaled.index = piv.index
        # piv = pivScaled
        
        # Perform log transformation (if specified)
        if(log_transform):
            piv[piv <= 0] = 0.00000001
            temp = piv.copy()
            piv = np.log10(piv)
            if(no_log !=None):
                for col in no_log:
                    piv[col] = temp[col]

        # Remove outliers
        if(remove_outliers):
            piv = preprocess.remove_outliers(piv, z_threshold=z_threshold)

        # Prepare data to plot
        samples = piv.shape[0]

        idx = piv.index.date
        dates = [dates.strftime('%Y-%m-%d') for dates in idx]
        remaining = [i for i in dates if i not in remove]
        piv = piv.loc[remaining]
        
        # Create the seaborn PairGrid plot
        sns.set_style("white", {"axes.facecolor": "0.95"})
        g = sns.PairGrid(piv, aspect=1.2, diag_sharey=False, despine=False)
        g.figure.suptitle(title, fontweight='bold', y=1.08, fontsize=25)
        g.map_lower(
            sns.regplot, 
            lowess=True, 
            ci=False, 
            line_kws={'color': 'red', 'lw': 3}, 
            scatter_kws={'color': 'black', 's': 20},
        )
        # g.map_diag(
        #     sns.distplot, 
        #     kde_kws={'color': 'black', 'lw': 3}, 
        #     hist_kws={'histtype': 'bar', 'lw': 2, 'edgecolor': 'k', 'facecolor':'grey'},
        # )   # NOTE: Replacing with `sns.histplot` because `sns.distplot` is deprecated.
        g.map_diag(
            sns.histplot, 
            stat="density",
            color="black",
            element="bars",
            edgecolor="dimgray",
            facecolor="lightgray",
            lw=2,
            kde=True, 
            kde_kws=dict(cut=3),
            line_kws=dict(color="black", lw=3),
        )

        g.map_upper(plots._plotUpperHalf)

        # for ax in g.axes.flat:
        #     ax.tick_params("y", labelrotation=0, labelsize=fontsize)
            
        #     # ax.set_xticklabels(
        #     #     ax.get_xticklabels(), rotation=45, fontsize=fontsize,
        #     # )   # NOTE: DISCOURAGED version of `xticks` - https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_xticklabels.html
        #     # ax.set_xticks(
        #     #     ax.get_xticks(),
        #     #     [f"{num:.2f}" for num in ax.get_xticks()],    # TODO: Check with Zexuan if this is supposed to be normalized.
        #     #     rotation=45, 
        #     #     fontsize=fontsize,
        #     # )     # NOTE: Non-discouraged version of `xticks`
        #     ax.tick_params("x", labelrotation=45, labelsize=fontsize)   # NOTE: Cleaner version of `xticks`
        #     ax.set_xlabel(
        #         ax.get_xlabel(), fontsize=fontsize, fontweight='bold',
        #     )
        #     ax.set_ylabel(
        #         ax.get_ylabel(), fontsize=fontsize, fontweight='bold',
        #     )
        _format_ticks_and_labels(axes=g.axes, fontsize=fontsize)
            
        # g.figure.subplots_adjust(wspace=0.3, hspace=0.3)
        g.figure.subplots_adjust(
            wspace=0.3, hspace=0.3, 
            left=0.1, right=0.75, 
            bottom=0.1, top=0.9,
        )

        # Add the meta-data information box
        ax = plt.gca()
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
        ax.text(
            1.3, 6.2, 
            # 'Start date:  {}\nEnd date:    {}\n\nOriginal samples:     {}\nSamples used:     {}'.format(piv.index[0].date(), piv.index[-1].date(), totalSamples, samples), 
            f"{'Start date:':<12} {str(piv.index[0].date())}\n{'End date:':<12} {str(piv.index[-1].date())}\n\n{'Original samples:':<20} {totalSamples}\n{'Samples used:':<20} {samples}",
            transform=ax.transAxes, 
            fontsize=fontsize, 
            fontweight='bold', 
            verticalalignment='bottom', 
            bbox=props,
        )
        
        # Add titles to the diagonal axes/subplots
        for ax, col in zip(np.diag(g.axes), piv.columns):
            ax.set_title(col, y=0.82, fontsize=15)
        
        # Save figure (if specified)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        g.figure.savefig(
            # save_dir + '/' + station_name + file_extension + '.png', 
            f"{save_dir}/{station_name}{file_extension}.png", 
            bbox_inches="tight",
        )
        
        # if(returnData):
        #     return piv
    
        return piv
    

def plot_all_corr_by_station(
        # self, 
        data_pylenm_dm,
        analytes, 
        remove_outliers=True, 
        z_threshold=4, 
        interpolate=False, 
        frequency='2W', 
        save_dir='plot_correlation', 
        log_transform=False, 
        fontsize=20,
    ):
    """Plots the correlations with the physical plots as station as the important analytes over time for each station in the dataset.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analytes (list): list of analyte names to use
        remove_outliers (bool, optional): choose whether or to remove the outliers. Defaults to True.
        z_threshold (int, optional): z_score threshold to eliminate outliers. Defaults to 4.
        interpolate (bool, optional): choose whether or to interpolate the data. Defaults to False.
        frequency (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} frequency to interpolate. Note: See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘W’ = every week, ‘D ’= every day, ‘2W’ = every 2 weeks). Defaults to '2W'.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_correlation'.
        log_transform (bool, optional): flag for log base 10 transformation. Defaults to False.
        fontsize (int, optional): font size. Defaults to 20.
    """
    
    # data = self.data
    data = data_pylenm_dm.data
    
    stations = np.array(data.STATION_ID.values)
    stations = np.unique(stations)
    
    for station in stations:
        plot_corr_by_station(
            data_pylenm_dm=data_pylenm_dm,
            station_name=station, 
            analytes=analytes,remove_outliers=remove_outliers, 
            z_threshold=z_threshold, 
            interpolate=interpolate, 
            frequency=frequency, 
            save_dir=save_dir, 
            log_transform=log_transform, 
            fontsize=fontsize,
        )

 
def plot_corr_by_date_range(
        # self, 
        data_pylenm_dm, 
        date, 
        analytes, 
        lag=0, 
        min_samples=10, 
        save_dir='plot_corr_by_date', 
        log_transform=False, 
        fontsize=20, 
        # returnData=False, 
        no_log=None,
    ):
    """Plots the correlations with the physical plots as station as the correlations of the important analytes for ALL the stations on a specified date or range of dates if a lag greater than 0 is specifed.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        date (str): date to be analyzed
        analytes (_type_): list of analyte names to use
        lag (int, optional): number of days to look ahead and behind the specified date (+/-). Defaults to 0.
        min_samples (int, optional): minimum number of samples the result should contain in order to execute.. Defaults to 10.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_corr_by_date'.
        log_transform (bool, optional): flag for log base 10 transformation. Defaults to False.
        fontsize (int, optional): font size. Defaults to 20.
        returnData (bool, optional): flag to return data used to perfrom correlation analysis. Defaults to False.
        no_log (list, optional): list of column names to not apply log transformation to. Defaults to None.
    """
    
    if (lag==0):
        
        # Prepare data
        data = data_pylenm_dm.data
        data = filters.simplify_data(data=data)
        query = data[data.COLLECTION_DATE == date]
        a = list(np.unique(query.ANALYTE_NAME.values))# get all analytes from dataset
        
        for value in analytes:
            # if((value in a)==False):
            if value not in a:
                plot_corr_logger.error(f"ERROR: No analyte named {value} in data.")
                # return 'ERROR: No analyte named "{}" in data.'.format(value)
                # return f"ERROR: No analyte named {value} in data."
                return None
        
        analytes = sorted(analytes)
        query = query.loc[query.ANALYTE_NAME.isin(analytes)]
        
        if(query.shape[0] == 0):
            plot_corr_logger.error(f"ERROR: {date} has no data for all of the analytes.")
            # return 'ERROR: {} has no data for all of the analytes.'.format(date)
            # return f"ERROR: {date} has no data for all of the analytes."
            return None
        
        samples = query[['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME']].duplicated().value_counts()[0]
        
        # Transform data if there are enough samples
        if(samples < min_samples):
            plot_corr_logger.error(f"ERROR: {date} does not have at least {min_samples} samples.")
            # return 'ERROR: {} does not have at least {} samples.'.format(date, min_samples)
            # return f"ERROR: {date} does not have at least {min_samples} samples."
            return None
        else:
            piv = query.reset_index().pivot_table(
                index='STATION_ID', 
                columns='ANALYTE_NAME', 
                values='RESULT', 
                aggfunc=np.mean,
            )
            # return piv
    
    else:
        # If the data has already been calculated with the lag specified, retrieve it
        # if(self.jointData_is_set(lag=lag)==True): 
        #     data = self.__jointData[0]
        if data_pylenm_dm.is_set_jointData(lag=lag): 
            data = data_pylenm_dm.jointData[0]
        
        # Otherwise, calculate it
        else:
            # data = self.getJointData(analytes, lag=lag)
            # self.__set_jointData(data=data, lag=lag)
            data = fetchers.getJointData(data_pylenm_dm=data_pylenm_dm, analytes=analytes, lag=lag)
            data_pylenm_dm.set_jointData(data=data, lag=lag)
        
        # Get new range based on the lag and create the pivot table to be able to do the correlation
        # dateStart, dateEnd = self.__getLagDate(date, lagDays=lag)
        dateStart, dateEnd = fetchers._getLagDate(date, lagDays=lag)
        dateRange_key = str(dateStart.date()) + " - " + str(dateEnd.date())
        piv = pd.DataFrame(data.loc[dateRange_key]).unstack().T
        piv.index = piv.index.droplevel()
        piv = pd.DataFrame(piv).dropna(axis=0, how='all')
        num_NaNs = int(piv.isnull().sum().sum())
        samples = (piv.shape[0] * piv.shape[1]) - num_NaNs
    
        # Convert columns to float and raise errors if the conversion fails
        for col in piv.columns:
            piv[col] = piv[col].astype('float64', errors='raise')
    
        if lag > 0:
            date = dateRange_key
        # return piv
    
    title = date + '_correlation'
    # scaler = StandardScaler()
    # pivScaled = scaler.fit_transform(piv)
    # pivScaled = pd.DataFrame(pivScaled, columns=piv.columns)
    # pivScaled.index = piv.index
    # piv = pivScaled

    # Perform log transform (if specified)
    if(log_transform):
        piv[piv <= 0] = 0.00000001
        temp = piv.copy()
        piv = np.log10(piv)
    
        if(no_log !=None):
            for col in no_log:
                piv[col] = temp[col]

    # Create seaborn PairGrid plot
    sns.set_style("white", {"axes.facecolor": "0.95"})
    g = sns.PairGrid(piv, aspect=1.2, diag_sharey=False, despine=False)
    g.figure.suptitle(title, fontweight='bold', y=1.08, fontsize=25)
    g.map_lower(
        sns.regplot, 
        lowess=True, 
        ci=False, 
        line_kws={'color': 'red', 'lw': 3}, 
        scatter_kws={'color': 'black', 's': 20},
    )
    # g.map_diag(
    #     sns.distplot, 
    #     kde_kws={'color': 'black', 'lw': 3}, 
    #     hist_kws={'histtype': 'bar', 'lw': 2, 'edgecolor': 'k', 'facecolor':'grey'},
    # )   # NOTE: Replacing with `sns.histplot` because `sns.distplot` is deprecated.
    g.map_diag(
        sns.histplot, 
        stat="density",
        color="black",
        element="bars",
        edgecolor="dimgray",
        facecolor="lightgray",
        lw=2,
        kde=True, 
        kde_kws=dict(cut=3),
        line_kws=dict(color="black", lw=3),
    )
    # g.map_upper(self.__plotUpperHalf)
    g.map_upper(plots._plotUpperHalf)
    
    # for ax in g.axes.flat:
    #     ax.tick_params("y", labelrotation=0, labelsize=fontsize)
    #     # ax.set_xticklabels(
    #     #     ax.get_xticklabels(), rotation=45, fontsize=fontsize,
    #     # )   # DISCOURAGED - https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_xticklabels.html
    #     ax.set_xticks(
    #         ax.get_xticks(),
    #         [f"{num:.2f}" for num in ax.get_xticks()],    # TODO: Check with Zexuan if this is supposed to be normalized.
    #         rotation=45, 
    #         fontsize=fontsize,
    #     )
    #     ax.set_xlabel(
    #         ax.get_xlabel(), fontsize=fontsize, fontweight='bold',
    #     ) #HERE
    #     ax.set_ylabel(
    #         ax.get_ylabel(), fontsize=fontsize, fontweight='bold',
    #     )
    _format_ticks_and_labels(axes=g.axes, fontsize=fontsize)

    # g.figure.subplots_adjust(wspace=0.3, hspace=0.3)
    g.figure.subplots_adjust(
        wspace=0.3, hspace=0.3, 
        left=0.1, right=0.75, 
        bottom=0.1, top=0.9,
    )
    ax = plt.gca()

    # Add the meta-data information box
    props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
    ax.text(
        1.3, 6.2, 
        f"Date: {date}\n\nStations: {piv.shape[0]}\nSamples used: {samples}",
        # 'Date:  {}\n\nStations:     {}\nSamples used:     {}'.format(date, piv.shape[0] ,samples), 
        transform=ax.transAxes, 
        fontsize=fontsize, 
        fontweight='bold', 
        verticalalignment='bottom', 
        bbox=props,
    )
    
    # Add titles to the diagonal axes/subplots
    for ax, col in zip(np.diag(g.axes), piv.columns):
        ax.set_title(col, y=0.82, fontsize=15)
    
    # Save figure (if specified)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    g.figure.savefig(
        # save_dir + '/' + date + '.png',
        f"{save_dir}/{date}.png", 
        bbox_inches="tight",
    )
    
    # if(returnData):
    #     return piv

    return piv


def plot_corr_by_year(
        # self, 
        data_pylenm_dm,
        year, 
        analytes, 
        remove_outliers=True, 
        z_threshold=4, 
        min_samples=10, 
        save_dir='plot_corr_by_year', 
        log_transform=False, 
        fontsize=20, 
        returnData=False, 
        no_log=None,
    ):
    """Plots the correlations with the physical plots as station as the correlations of the important analytes for ALL the stations in specified year.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        year (int): year to be analyzed
        analytes (list): list of analyte names to use
        remove_outliers (bool, optional): choose whether or to remove the outliers.. Defaults to True.
        z_threshold (int, optional): z_score threshold to eliminate outliers. Defaults to 4.
        min_samples (int, optional): minimum number of samples the result should contain in order to execute.. Defaults to 10.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_correlation'.
        log_transform (bool, optional): flag for log base 10 transformation. Defaults to False.
        fontsize (int, optional): font size. Defaults to 20.
        returnData (bool, optional): flag to return data used to perfrom correlation analysis. Defaults to False.
        no_log (list, optional): list of column names to not apply log transformation to. Defaults to None.
    """
    
    # Prepare data
    # data = self.data
    data = data_pylenm_dm.data
    query = data
    # query = self.simplify_data(data=query)
    query = filters.simplify_data(data=query)
    query.COLLECTION_DATE = pd.to_datetime(query.COLLECTION_DATE)
    query = query[query.COLLECTION_DATE.dt.year == year]
    a = list(np.unique(query.ANALYTE_NAME.values))# get all analytes from dataset
    
    for value in analytes:
        if((value in a)==False):
            plot_corr_logger.error(f"ERROR: No analyte named {value} in data.")
            # return 'ERROR: No analyte named "{}" in data.'.format(value)
            return None
    
    analytes = sorted(analytes)
    
    # Performing checks for enough data for the plots
    query = query.loc[query.ANALYTE_NAME.isin(analytes)]
    if(query.shape[0] == 0):
        plot_corr_logger.error(f"ERROR: {year} has no data for the 6 analytes.")
        # return 'ERROR: {} has no data for the 6 analytes.'.format(year)
        return None

    samples = query[['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME']].duplicated().value_counts()[0]
    if(samples < min_samples):
        plot_corr_logger.error(f"ERROR: {year} does not have at least {min_samples} samples.")
        # return 'ERROR: {} does not have at least {} samples.'.format(date, min_samples)
        return None

    else:
        # Create pivot table if there is enough data
        piv = query.reset_index().pivot_table(
            index='STATION_ID', 
            columns='ANALYTE_NAME', 
            values='RESULT', 
            aggfunc=np.mean,
        )
        # return piv
        
        # Remove outliers
        if(remove_outliers):
            # piv = self.remove_outliers(piv, z_threshold=z_threshold)
            piv = preprocess.remove_outliers(piv, z_threshold=z_threshold)
        
        samples = piv.shape[0] * piv.shape[1]

        title = str(year) + '_correlation'
        # scaler = StandardScaler()
        # pivScaled = scaler.fit_transform(piv)
        # pivScaled = pd.DataFrame(pivScaled, columns=piv.columns)
        # pivScaled.index = piv.index
        # piv = pivScaled

        # Perform log transformation (if specified)
        if(log_transform):
            piv[piv <= 0] = 0.00000001
            temp = piv.copy()
            piv = np.log10(piv)
            
            if(no_log !=None):
                for col in no_log:
                    piv[col] = temp[col]

        # Create the seaborn PairGrid plot
        sns.set_style("white", {"axes.facecolor": "0.95"})
        g = sns.PairGrid(piv, aspect=1.2, diag_sharey=False, despine=False)
        g.figure.suptitle(title, fontweight='bold', y=1.08, fontsize=25)
        g.map_lower(
            sns.regplot, 
            lowess=True, 
            ci=False, 
            line_kws={'color': 'red', 'lw': 3}, 
            scatter_kws={'color': 'black', 's': 20},
        )
        # g.map_diag(
        #     sns.distplot, 
        #     kde_kws={'color': 'black', 'lw': 3}, 
        #     hist_kws={'histtype': 'bar', 'lw': 2, 'edgecolor': 'k', 'facecolor':'grey'},
        # )   # NOTE: Replacing with `sns.histplot` because `sns.distplot` is deprecated.
        g.map_diag(
            sns.histplot, 
            stat="density",
            color="black",
            element="bars",
            edgecolor="dimgray",
            facecolor="lightgray",
            lw=2,
            kde=True, 
            kde_kws=dict(cut=3),
            line_kws=dict(color="black", lw=3),
        )
        # g.map_upper(self.__plotUpperHalf)
        g.map_upper(plots._plotUpperHalf)
        
        # for ax in g.axes.flat:
        #     ax.tick_params("y", labelrotation=0, labelsize=fontsize)
        #     # ax.set_xticklabels(
        #     #     ax.get_xticklabels(), rotation=45, fontsize=fontsize,
        #     # )   # DISCOURAGED - https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_xticklabels.html
        #     ax.set_xticks(
        #         ax.get_xticks(),
        #         [f"{num:.2f}" for num in ax.get_xticks()],    # TODO: Check with Zexuan if this is supposed to be normalized.
        #         rotation=45, 
        #         fontsize=fontsize,
        #     )
        #     ax.set_xlabel(
        #         ax.get_xlabel(), fontsize=fontsize, fontweight='bold',
        #     )
        #     ax.set_ylabel(
        #         ax.get_ylabel(), fontsize=fontsize, fontweight='bold',
        #     )
        _format_ticks_and_labels(axes=g.axes, fontsize=fontsize)

        # g.figure.subplots_adjust(wspace=0.3, hspace=0.3)
        g.figure.subplots_adjust(
            wspace=0.3, hspace=0.3, 
            left=0.1, right=0.75, 
            bottom=0.1, top=0.9,
        )

        # Add the meta-data information box
        ax = plt.gca()
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
        ax.text(
            1.3, 
            # 3, 
            6.2, 
            # 'Date:  {}\n\nSamples used:     {}'.format(year, samples), 
            f"Year: {year}\n\nSamples used: {samples}",
            transform=ax.transAxes, 
            fontsize=fontsize, 
            fontweight='bold', 
            verticalalignment='bottom', 
            bbox=props,
        )
        
        # Add titles to the diagonal axes/subplots
        for ax, col in zip(np.diag(g.axes), piv.columns):
            ax.set_title(col, y=0.82, fontsize=15)
        
        # Save figure (if specified)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        g.figure.savefig(save_dir + '/' + str(year) + '.png', bbox_inches="tight")
        
        # if(returnData):
        #     return piv

        return piv


import numpy as np
import pandas as pd

from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

df_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


# class PylenmDataFactory(object):
class PylenmDataModule(object):
    """Class object that initilaizes Pylenm given data.
    """
    
    def __init__(
        self, 
        data: pd.DataFrame=None,
        construction_data: pd.DataFrame=None,
        logger_level:int=logging.WARNING,
    ) -> None:
        """Initializes pylenm with a Pandas DataFrame

        Args:
            data (pd.DataFrame): Data to be imported.
        """

        self.logger = logger_config.setup_logging(module_name=__name__, level=logger_level)

        if data is not None: 
            self.set_data(data, verbose=True)
        else:
            self.data = None
        
        if construction_data is not None:
            self.set_construction_data(construction_data, verbose=True)
        else:
            self.construction_data = None
        
        self.__jointData = [None, 0]    # TODO: Check if we can make this a single value


    def __has_columns(self, data: pd.DataFrame, required_cols=list()):
        data_cols = set([x.upper() for x in data.columns])
        return set(required_cols).issubset(data_cols)


    def __is_valid(self, data: pd.DataFrame, required_cols=list()):
        """Validates data for Pylenm usage.

        Args:
            data (pd.DataFrame): Data to be validated.
            required_cols (list, optional): List of required columns. Defaults to REQUIRED_DATA_COLUMNS.

        Returns:
            bool: True if data is valid, False otherwise.
        """

        if data is None:
            self.logger.error("`data` is None!")
            return False
        
        if not isinstance(data, pd.DataFrame):
            self.logger.error("`data` must be a Pandas DataFrame!")
            return False
        
        if not self.__has_columns(data=data, required_cols=required_cols):
            missing_cols = list(
                set(required_cols).difference(set(data.columns))
            )
            self.logger.error(f"Missing these columns from the data: {missing_cols}.")
            return False

        return True


    def is_valid_data(self, data: pd.DataFrame):
        return self.__is_valid(data, required_cols=c.REQUIRED_DATA_COLUMNS)
    

    def is_valid_construction_data(self, data: pd.DataFrame):
        return self.__is_valid(data, required_cols=c.REQUIRED_CONSTRUCTION_DATA_COLUMNS)

    
    def __set_units(self):
        # analytes = list(np.unique(self.data[['ANALYTE_NAME']]))
        # mask1 = ~self.data[['ANALYTE_NAME','RESULT_UNITS']].duplicated()
        # res = self.data[['ANALYTE_NAME','RESULT_UNITS']][mask1]
        # mask2 = ~self.data[['ANALYTE_NAME']].duplicated()
        # res = res[mask2]        # NOTE: Generates warning => UserWarning: Boolean Series key will be reindexed to match DataFrame index. # TODO: Check later.
        
        res = self.data[['ANALYTE_NAME','RESULT_UNITS']].drop_duplicates(
            subset="ANALYTE_NAME",
        )
        unit_dictionary = pd.Series(res.RESULT_UNITS.values,index=res.ANALYTE_NAME).to_dict()
        self.unit_dictionary = unit_dictionary
    

    def get_unit(self, analyte_name):
        """Returns the unit of the analyte you specify. Example: 'DEPTH_TO_WATER' may return 'ft'

        Args:
            analyte_name (str): ame of the analyte to be processed

        Returns:
            str: unit of analyte
        """
        return self.unit_dictionary[analyte_name]


    # SETTING DATA
    def set_data(self, data: pd.DataFrame, verbose: bool=False) -> None:
        """Saves the dataset into pylenm.

        Args:
            data (pd.DataFrame): Dataset to be imported.
            verbose (bool, optional): Prints success message. Defaults to True.

        Returns:
            None
        """
        if self.is_valid_data(data):
            cols_upper = [x.upper() for x in data.columns]
            data.columns = cols_upper
            self.data = data
            self.logger.info("Successfully imported the data!")
            
            if(verbose):
                print('Successfully imported the data!\n')
            
            self.__set_units()
        
        else:
            self.logger.info("Failed to import data!")
            self.logger.warning(c.PYLENM_DATA_REQUIREMENTS)


    def set_construction_data(
        self, 
        construction_data: pd.DataFrame, 
        verbose: bool=False,
    ) -> None:
        """Imports the additional station information as a separate DataFrame.

        Args:
            construction_data (pd.DataFrame): Data with additonal details.
            verbose (bool, optional): Prints success message. Defaults to True.

        Returns:
            None
        """
        if self.is_valid_construction_data(construction_data):
            cols_upper = [x.upper() for x in list(construction_data.columns)]
            construction_data.columns = cols_upper
            self.construction_data = construction_data.set_index(['STATION_ID'])
            self.logger.info("Successfully imported the construction data!")
            
            if(verbose):
                print('Successfully imported the construction data!\n')
        
        else:
            self.logger.info("Failed to import the construction data!")
            self.logger.warning(c.PYLENM_CONSTRUCTION_DATA_REQUIREMENTS)


    def set_jointData(self, data, lag):
        self.__jointData[0] = data
        self.__jointData[1] = lag

    
    @property
    def jointData(self):
        return self.__jointData
    

    # def jointData_is_set(self, lag):
    def is_set_jointData(self, lag):
        """Checks to see if getJointData function was already called and saved for given lag.

        Args:
            lag (int): number of days to look ahead and behind the specified date (+/-)

        Returns:
            bool: True if JointData was already calculated, False, otherwise.
        """
        # if(str(type(self.__jointData[0])).lower().find('dataframe') == -1):
        if not isinstance(self.__jointData[0], pd.DataFrame):
            return False
        
        if self.__jointData[1] != lag:
            return False
        
        return True


    def get_data(self):
        return self.data
    

    def get_construction_data(self):
        return self.construction_data


from math import sqrt
from sklearn.metrics import mean_squared_error


def dist(p1, p2):
    """2D Euclidean distance function

    Args:
        p1 (tuple): first point
        p2 (tuple): second point

    Returns:
        float: Euclidean distance
    """
    return sqrt(((p1[0]-p2[0])**2)+((p1[1]-p2[1])**2))


def mse(y_true, y_pred):
    """Error Metric: Mean Squared Error 

    Args:
        y_true (numpy.array): true values
        y_pred (numpy.array): predicted values

    Returns:
        float: mean squared error
    """
    return mean_squared_error(y_true, y_pred)